from __future__ import annotations

import contextlib
from typing import Any, List

from pyinstrument import processors
from pyinstrument.frame import Frame
from pyinstrument.session import Session

# pyright: strict


ProcessorList = List[processors.ProcessorType]


class Renderer:
    """
    Abstract base class for renderers.
    """

    output_file_extension: str = "txt"
    """
    Renderer output file extension without dot prefix. The default value is `txt`
    """

    output_is_binary: bool = False
    """
    Whether the output of this renderer is binary data. The default value is `False`.
    """

    def __init__(self):
        pass

    def render(self, session: Session) -> str:
        """
        Return a string that contains the rendered form of `frame`.
        """
        raise NotImplementedError()

    class MisconfigurationError(Exception):
        pass


class FrameRenderer(Renderer):
    """
    An abstract base class for renderers that process Frame objects using
    processor functions. Provides a common interface to manipulate the
    processors before rendering.
    """

    processors: ProcessorList
    """
    Processors installed on this renderer. This property is defined on the
    base class to provide a common way for users to add and
    manipulate them before calling :func:`render`.
    """

    processor_options: dict[str, Any]
    """
    Dictionary containing processor options, passed to each processor.
    """

    show_all: bool
    timeline: bool

    def __init__(
        self,
        show_all: bool = False,
        timeline: bool = False,
        processor_options: dict[str, Any] | None = None,
    ):
        """
        :param show_all: Don't hide or filter frames - show everything that pyinstrument captures.
        :param timeline: Instead of aggregating time, leave the samples in chronological order.
        :param processor_options: A dictionary of processor options.
        """
        # processors is defined on the base class to provide a common way for users to
        # add to and manipulate them before calling render()
        self.processors = self.default_processors()
        self.processor_options = processor_options or {}

        self.show_all = show_all
        self.timeline = timeline

        if show_all:
            for p in (
                processors.group_library_frames_processor,
                processors.remove_importlib,
                processors.remove_irrelevant_nodes,
                processors.remove_tracebackhide,
                # note: we're not removing these processors
                # processors.remove_unnecessary_self_time_nodes,
                #    (still hide the inner pyinstrument synthetic frames)
                # processors.remove_first_pyinstrument_frames_processor,
                #    (still hide the outer pyinstrument calling frames)
            ):
                with contextlib.suppress(ValueError):
                    # don't care if the processor isn't in the list
                    self.processors.remove(p)

        if timeline:
            with contextlib.suppress(ValueError):
                self.processors.remove(processors.aggregate_repeated_calls)

    def default_processors(self) -> ProcessorList:
        """
        Return a list of processors that this renderer uses by default.
        """
        raise NotImplementedError()

    def preprocess(self, root_frame: Frame | None) -> Frame | None:
        frame = root_frame
        for processor in self.processors:
            frame = processor(frame, options=self.processor_options)
        return frame

    def render(self, session: Session) -> str:
        """
        Return a string that contains the rendered form of `frame`.
        """
        raise NotImplementedError()



from __future__ import annotations

import fnmatch
import glob
import json
import optparse
import os
import runpy
import shutil
import sys
import time
from typing import Any, List, TextIO, cast

import pyinstrument
from pyinstrument import Profiler, renderers
from pyinstrument.session import Session
from pyinstrument.util import (
    file_is_a_tty,
    file_supports_color,
    file_supports_unicode,
    object_with_import_path,
)
from pyinstrument.vendor import appdirs, keypath

# pyright: strict
# pyright: reportUnknownMemberType=false


def main():
    usage = "usage: pyinstrument [options] scriptfile [arg] ..."
    version_string = "pyinstrument {v}, on Python {pyv[0]}.{pyv[1]}.{pyv[2]}".format(
        v=pyinstrument.__version__,
        pyv=sys.version_info,
    )
    parser: Any = optparse.OptionParser(usage=usage, version=version_string)
    parser.allow_interspersed_args = False

    def store_and_consume_remaining(
        option: optparse.Option, opt: str, value: str, parser: optparse.OptionParser
    ):
        """
        A callback for optparse that stores the value and consumes all
        remaining arguments, storing them in the same variable as a tuple.
        """

        # assert a few things we know to be true about the parser
        assert option.dest
        assert parser.rargs is not None
        assert parser.largs is not None

        # everything after this argument should be consumed
        remaining_arguments = parser.rargs + parser.largs
        parser.rargs[:] = []
        parser.largs[:] = []

        setattr(parser.values, option.dest, ValueWithRemainingArgs(value, remaining_arguments))

    parser.add_option(
        "--load",
        dest="load",
        action="store",
        metavar="FILENAME",
        help="instead of running a script, load a profile session from a pyisession file",
    )

    parser.add_option(
        "",
        "--load-prev",
        dest="load_prev",
        action="store",
        metavar="IDENTIFIER",
        help="instead of running a script, load a previous profile session as specified by an identifier",
    )

    parser.add_option(
        "-m",
        "",
        dest="module",
        action="callback",
        callback=store_and_consume_remaining,
        type="string",
        help="run library module as a script, like 'python -m module'",
    )
    parser.add_option(
        "-c",
        "",
        dest="program",
        action="callback",
        callback=store_and_consume_remaining,
        type="string",
        help="program passed in as string, like 'python -c \"...\"'",
    )
    parser.add_option(
        "",
        "--from-path",
        dest="from_path",
        action="store_true",
        help="(POSIX only) instead of the working directory, look for scriptfile in the PATH environment variable",
    )

    parser.add_option(
        "-o", "--outfile", dest="outfile", action="store", help="save to <outfile>", default=None
    )

    parser.add_option(
        "-r",
        "--renderer",
        dest="renderer",
        action="store",
        type="string",
        help=(
            "how the report should be rendered. One of: 'text', 'html', 'json', 'speedscope', "
            "'pyisession', 'pstats', or python import path to a renderer class. Defaults to "
            "the appropriate format for the extension if OUTFILE is given, otherwise, defaults "
            "to 'text'."
        ),
        default=None,
    )

    parser.add_option(
        "-p",
        "--render-option",
        dest="render_options",
        action="append",
        metavar="RENDER_OPTION",
        type="string",
        help=(
            "options to pass to the renderer, in the format 'flag_name' or 'option_name=option_value'. "
            "For example, to set the option 'time', pass '-p time=percent_of_total'. To pass multiple "
            "options, use the -p option multiple times. You can set processor options using dot-syntax, "
            "like '-p processor_options.filter_threshold=0'. option_value is parsed as a JSON value or "
            "a string."
        ),
    )

    parser.add_option(
        "",
        "--html",
        dest="output_html",
        action="store_true",
        help=optparse.SUPPRESS_HELP,
        default=False,
    )  # deprecated shortcut for --renderer=html

    parser.add_option(
        "-t",
        "--timeline",
        dest="timeline",
        action="store_true",
        default=False,
        help="render as a timeline - preserve ordering and don't condense repeated calls",
    )

    parser.add_option(
        "",
        "--hide",
        dest="hide_fnmatch",
        action="store",
        metavar="EXPR",
        help=(
            "glob-style pattern matching the file paths whose frames to hide. Defaults to "
            "hiding non-application code"
        ),
    )
    parser.add_option(
        "",
        "--hide-regex",
        dest="hide_regex",
        action="store",
        metavar="REGEX",
        help=(
            "regex matching the file paths whose frames to hide. Useful if --hide doesn't give "
            "enough control."
        ),
    )

    parser.add_option(
        "",
        "--show",
        dest="show_fnmatch",
        action="store",
        metavar="EXPR",
        help=(
            "glob-style pattern matching the file paths whose frames to "
            "show, regardless of --hide or --hide-regex. For example, use "
            "--show '*/<library>/*' to show frames within a library that "
            "would otherwise be hidden."
        ),
    )
    parser.add_option(
        "",
        "--show-regex",
        dest="show_regex",
        action="store",
        metavar="REGEX",
        help=(
            "regex matching the file paths whose frames to always show. "
            "Useful if --show doesn't give enough control."
        ),
    )
    parser.add_option(
        "",
        "--show-all",
        dest="show_all",
        action="store_true",
        help="show everything",
        default=False,
    )

    parser.add_option(
        "",
        "--unicode",
        dest="unicode",
        action="store_true",
        help="(text renderer only) force unicode text output",
    )
    parser.add_option(
        "",
        "--no-unicode",
        dest="unicode",
        action="store_false",
        help="(text renderer only) force ascii text output",
    )

    parser.add_option(
        "",
        "--color",
        dest="color",
        action="store_true",
        help="(text renderer only) force ansi color text output",
    )
    parser.add_option(
        "",
        "--no-color",
        dest="color",
        action="store_false",
        help="(text renderer only) force no color text output",
    )
    parser.add_option(
        "-i",
        "--interval",
        action="store",
        type=float,
        help=(
            "Minimum time, in seconds, between each stack sample. Smaller values "
            "allow resolving shorter duration function calls but incur a "
            "greater runtime and memory consumption overhead. For longer running "
            "scripts, setting a larger interval reduces the memory consumption "
            "required to store the stack samples."
        ),
        default=0.001,
    )
    parser.add_option(
        "",
        "--use-timing-thread",
        dest="use_timing_thread",
        action="store_true",
        help=(
            "Use a separate thread to time the interval between stack samples. "
            "This can reduce the overhead of sampling on some systems."
        ),
    )

    # parse the options

    if not sys.argv[1:]:
        parser.print_help()
        sys.exit(2)

    options, args = parser.parse_args()  # type: ignore

    # make command line options type-checked
    options = cast(CommandLineOptions, options)
    # work around a type checking bug...
    args = cast(List[str], args)  # type: ignore

    session_options_used = [
        options.load is not None,
        options.load_prev is not None,
        options.module is not None,
        options.program is not None,
        len(args) > 0,
    ]
    if session_options_used.count(True) == 0:
        parser.print_help()
        sys.exit(2)
    if session_options_used.count(True) > 1:
        parser.error("You can only specify one of --load, --load-prev, -m, or script arguments")

    if options.module is not None and options.from_path:
        parser.error("The options -m and --from-path are mutually exclusive.")

    if options.from_path and sys.platform == "win32":
        parser.error("--from-path is not supported on Windows")

    renderer_class = get_renderer_class(options)

    # open the output file

    if options.outfile:
        f = open(
            options.outfile,
            "w",
            encoding="utf-8",
            errors="surrogateescape",
            newline="" if renderer_class.output_is_binary else None,
        )
        should_close_f_after_writing = True
    else:
        f = sys.stdout
        should_close_f_after_writing = False

    inner_exception = None

    # create the renderer

    try:
        renderer = create_renderer(renderer_class, options, output_file=f)
    except OptionsParseError as e:
        parser.error(e.args[0])
        exit(1)

    if renderer.output_is_binary and not options.outfile and file_is_a_tty(f):
        parser.error(
            "Can't write binary output to a terminal. Redirect to a file or use --outfile."
        )
        exit(1)

    # get the session - execute code or load from disk

    if options.load_prev:
        session = load_report_from_temp_storage(options.load_prev)
    elif options.load:
        session = Session.load(options.load)
    else:
        # we are running some code
        if options.module is not None:
            if not (sys.path[0] and os.path.samefile(sys.path[0], ".")):
                # when called with '-m', search the cwd for that module
                sys.path[0] = os.path.abspath(".")

            argv = [options.module.value] + options.module.remaining_args
            code = "run_module(modname, run_name='__main__', alter_sys=True)"
            globs = {"run_module": runpy.run_module, "modname": options.module.value}
        elif options.program is not None:
            argv = ["-c", *options.program.remaining_args]
            code = options.program.value
            globs = {"__name__": "__main__"}
            # set the first path entry to '' to match behaviour of python -c
            sys.path[0] = ""
        else:
            argv = args
            if options.from_path:
                progname = shutil.which(args[0])
                if progname is None:
                    sys.exit(f"Error: program {args[0]} not found in PATH!")
            else:
                progname = args[0]
                if not os.path.exists(progname):
                    sys.exit(f"Error: program {args[0]} not found!")

            # Make sure we overwrite the first entry of sys.path ('.') with directory of the program.
            sys.path[0] = os.path.dirname(progname)

            code = "run_path(progname, run_name='__main__')"
            globs = {"run_path": runpy.run_path, "progname": progname}

        old_argv = sys.argv.copy()

        # there is no point using async mode for command line invocation,
        # because it will always be capturing the whole program, we never want
        # any execution to be <out-of-context>, and it avoids duplicate
        # profiler errors.
        profiler = Profiler(
            interval=options.interval,
            async_mode="disabled",
            use_timing_thread=options.use_timing_thread,
        )

        profiler.start(target_description=f'Program: {" ".join(argv)}')

        try:
            sys.argv[:] = argv
            exec(code, globs, None)
        except (SystemExit, KeyboardInterrupt) as e:
            inner_exception = e
        finally:
            sys.argv[:] = old_argv

        session = profiler.stop()

    if isinstance(renderer, renderers.HTMLRenderer) and not options.outfile and file_is_a_tty(f):
        # don't write HTML to a TTY, open in browser instead
        output_filename = renderer.open_in_browser(session)
        print("stdout is a terminal, so saved profile output to %s" % output_filename)
    else:
        f.write(renderer.render(session))
        if should_close_f_after_writing:
            f.close()

    if isinstance(renderer, renderers.ConsoleRenderer) and not options.outfile:
        _, report_identifier = save_report_to_temp_storage(session)
        print("To view this report with different options, run:")
        print("    pyinstrument --load-prev %s [options]" % report_identifier)
        print("")

    if inner_exception:
        # If the script raised an exception, re-raise it now to resume
        # the normal Python exception handling (printing the traceback, etc.)
        raise inner_exception


class OptionsParseError(Exception):
    pass


def compute_render_options(
    options: CommandLineOptions,
    renderer_class: type[renderers.Renderer],
    unicode_support: bool,
    color_support: bool,
) -> dict[str, Any]:
    """
    Given a list of `CommandLineOptions`, compute the
    rendering options for the given renderer.

    Raises an `OptionsParseError` if there is an error parsing the options.

    unicode_support:
        indicate whether the expected output supports unicode
    color_support:
        indicate whether the expected output supports color

    Both of these will be used to determine the  default of outputting unicode
    or color, but can be overridden with `options.color` and `option.unicode`.
    """

    # parse show/hide options
    if options.hide_fnmatch is not None and options.hide_regex is not None:
        raise OptionsParseError("You can‘t specify both --hide and --hide-regex")

    hide_regex: str | None
    show_regex: str | None
    if options.hide_fnmatch is not None:
        hide_regex = fnmatch.translate(options.hide_fnmatch)
    else:
        hide_regex = options.hide_regex

    show_options_used = [
        options.show_fnmatch is not None,
        options.show_regex is not None,
        options.show_all,
    ]
    if show_options_used.count(True) > 1:
        raise OptionsParseError("You can only specify one of --show, --show-regex and --show-all")

    if options.show_fnmatch is not None:
        show_regex = fnmatch.translate(options.show_fnmatch)
    elif options.show_all:
        show_regex = r".*"
    else:
        show_regex = options.show_regex

    render_options: dict[str, Any] = {}

    if issubclass(renderer_class, renderers.FrameRenderer):
        render_options["processor_options"] = {
            "hide_regex": hide_regex,
            "show_regex": show_regex,
        }

    if issubclass(renderer_class, renderers.ConsoleRenderer):
        unicode_override = options.unicode is not None
        color_override = options.color is not None
        unicode: Any = options.unicode if unicode_override else unicode_support
        color: Any = options.color if color_override else color_support

        render_options.update({"unicode": unicode, "color": color})

    if options.timeline:
        render_options["timeline"] = True
    if options.show_all:
        render_options["show_all"] = True

    # apply user options
    if options.render_options is not None:
        for renderer_option in options.render_options:
            key, sep, value = renderer_option.partition("=")

            if sep == "":
                # we're setting a flag, like `-p unicode`
                keypath.set_value_at_keypath(render_options, key, True)
            else:
                # it's a key=value structure
                try:
                    # try parsing as a JSON value
                    parsed_value = json.loads(value)
                except json.JSONDecodeError:
                    # otherwise treat it as a string
                    parsed_value = value

                keypath.set_value_at_keypath(render_options, key, parsed_value)

    return render_options


def create_renderer(
    renderer_class: type[renderers.Renderer], options: CommandLineOptions, output_file: TextIO
) -> renderers.Renderer:
    render_options = compute_render_options(
        options,
        renderer_class=renderer_class,
        unicode_support=file_supports_unicode(output_file),
        color_support=file_supports_color(output_file),
    )

    try:
        return renderer_class(**render_options)
    except (TypeError, renderers.Renderer.MisconfigurationError) as err:
        # TypeError is probably a bad renderer option, so we produce a nicer error message
        raise OptionsParseError(
            f"Failed to create {renderer_class.__name__}. Check your renderer options.\n  {err}\n"
        )


def get_renderer_class(options: CommandLineOptions) -> type[renderers.Renderer]:
    renderer = options.renderer

    if options.output_html:
        renderer = "html"

    if renderer is None and options.outfile:
        renderer = guess_renderer_from_outfile(options.outfile)

    if renderer is None:
        renderer = "text"

    if renderer == "text":
        return renderers.ConsoleRenderer
    elif renderer == "html":
        return renderers.HTMLRenderer
    elif renderer == "json":
        return renderers.JSONRenderer
    elif renderer == "speedscope":
        return renderers.SpeedscopeRenderer
    elif renderer == "pyisession" or renderer == "session":  # session is the old name
        return renderers.SessionRenderer
    elif renderer == "pstats":
        return renderers.PstatsRenderer
    else:
        try:
            return object_with_import_path(renderer)
        except (ValueError, ModuleNotFoundError, AttributeError) as err:
            # ValueError means we failed to import this object
            raise OptionsParseError(
                f"Failed to find renderer with name {renderer!r}.\n"
                "Options are text, html, json, speedscope, pstats or a Python\n"
                "import path to a Renderer class.\n"
                "\n"
                f"Underlying error: {err}\n"
            )


def guess_renderer_from_outfile(outfile: str) -> str | None:
    # ignore case of outfile
    outfile = outfile.lower()

    _, ext = os.path.splitext(outfile)

    if ext == ".txt":
        return "text"
    elif ext in [".html", ".htm"]:
        return "html"
    elif outfile.endswith(".speedscope.json"):
        return "speedscope"
    elif ext == ".json":
        return "json"
    elif ext == ".pyisession":
        return "session"
    elif ext == ".pstats":
        return "pstats"
    else:
        return None


def report_dir() -> str:
    data_dir = appdirs.user_data_dir("pyinstrument", "com.github.joerick")  # type: ignore
    report_dir = os.path.join(data_dir, "reports")
    if not os.path.exists(report_dir):
        os.makedirs(report_dir)
    return report_dir


def load_report_from_temp_storage(identifier: str) -> Session:
    """
    Returns the session referred to by identifier
    """
    path = os.path.join(report_dir(), identifier + ".pyisession")
    try:
        return Session.load(path)
    except FileNotFoundError:
        sys.exit(f"pyinstrument: Couldn't find a profile with identifier {identifier}")


def save_report_to_temp_storage(session: Session):
    """
    Saves the session to a temp file, and returns that path.
    Also prunes the number of reports to 10 so there aren't loads building up.
    """
    # prune this folder to contain the last 10 sessions
    previous_reports = glob.glob(os.path.join(report_dir(), "*.pyisession"))
    previous_reports.sort(reverse=True)
    while len(previous_reports) > 10:
        report_file = previous_reports.pop()
        os.remove(report_file)

    identifier = time.strftime("%Y-%m-%dT%H-%M-%S", time.localtime(session.start_time))

    path = os.path.join(report_dir(), identifier + ".pyisession")
    session.save(path)
    return path, identifier


class CommandLineOptions:
    """
    A type that codifies the `options` variable.
    """

    module: ValueWithRemainingArgs | None
    program: ValueWithRemainingArgs | None
    load: str | None
    load_prev: str | None
    from_path: str | None
    hide_fnmatch: str | None
    show_fnmatch: str | None
    hide_regex: str | None
    show_regex: str | None
    show_all: bool
    output_html: bool
    outfile: str | None
    render_options: list[str] | None

    unicode: bool | None
    color: bool | None
    renderer: str | None
    timeline: bool
    interval: float
    use_timing_thread: bool | None


class ValueWithRemainingArgs:
    def __init__(self, value: str, remaining_args: list[str]):
        self.value = value
        self.remaining_args = remaining_args


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import time
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Union

from pyinstrument import processors
from pyinstrument.frame import Frame
from pyinstrument.renderers.base import FrameRenderer, ProcessorList
from pyinstrument.session import Session

# pyright: strict


@dataclass(frozen=True, eq=True)
class SpeedscopeFrame:
    """
    Data class to store data needed for speedscope's concept of a
    frame, hereafter referred to as a "speedscope frame", as opposed to
    a "pyinstrument frame". This type must be hashable in order to use
    it as a dictionary key; a dictionary will be used to track unique
    speedscope frames.
    """

    name: str | None
    file: str | None
    line: int | None


class SpeedscopeEventType(Enum):
    """Enum representing the only two types of speedscope frame events"""

    OPEN = "O"
    CLOSE = "C"


@dataclass
class SpeedscopeEvent:
    """
    Data class to store speedscope's concept of an "event", which
    corresponds to opening or closing stack frames as functions or
    methods are entered or exited.
    """

    type: SpeedscopeEventType
    at: float
    frame: int


@dataclass
class SpeedscopeProfile:
    """
    Data class to store speedscope's concept of a "profile".
    """

    name: str
    events: list[SpeedscopeEvent]
    end_value: float
    start_value: float = 0.0
    type: str = "evented"
    unit: str = "seconds"


@dataclass
class SpeedscopeFile:
    """
    Data class encoding fields in speedscope's JSON file schema
    """

    name: str
    profiles: list[SpeedscopeProfile]
    shared: dict[str, list[SpeedscopeFrame]]
    schema: str = "https://www.speedscope.app/file-format-schema.json"
    active_profile_index: None = None
    exporter: str = "pyinstrument"


SpeedscopeFrameDictType = Dict[str, Union[str, int, None]]
SpeedscopeEventDictType = Dict[str, Union[SpeedscopeEventType, float, int]]


class SpeedscopeEncoder(json.JSONEncoder):
    """
    Encoder class used by json.dumps to serialize the various
    speedscope data classes.
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, SpeedscopeFile):
            return {
                "$schema": o.schema,
                "name": o.name,
                "activeProfileIndex": o.active_profile_index,
                "exporter": o.exporter,
                "profiles": o.profiles,
                "shared": o.shared,
            }
        if isinstance(o, SpeedscopeProfile):
            return {
                "type": o.type,
                "name": o.name,
                "unit": o.unit,
                "startValue": o.start_value,
                "endValue": o.end_value,
                "events": o.events,
            }
        if isinstance(o, (SpeedscopeFrame, SpeedscopeEvent)):
            d: SpeedscopeFrameDictType | SpeedscopeEventDictType = o.__dict__
            return d
        if isinstance(o, SpeedscopeEventType):
            return o.value
        return json.JSONEncoder.default(self, o)


class SpeedscopeRenderer(FrameRenderer):
    """
    Outputs a tree of JSON conforming to the speedscope schema documented at

    wiki: https://github.com/jlfwong/speedscope/wiki/Importing-from-custom-sources
    schema: https://www.speedscope.app/file-format-schema.json
    spec: https://github.com/jlfwong/speedscope/blob/main/src/lib/file-format-spec.ts
    example: https://github.com/jlfwong/speedscope/blob/main/sample/profiles/speedscope/0.0.1/simple.speedscope.json

    """

    output_file_extension = "speedscope.json"

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

        # Member holding a running total of wall clock time needed to
        # compute the times at which events occur
        self._event_time: float = 0.0

        # Map of speedscope frames to speedscope frame indices, needed
        # to construct evented speedscope profiles; exploits LIFO
        # property of popinfo method in Python 3.7+ dictionaries. This
        # dictionary is used to build up the "shared" JSON array in
        # speedscope's schema.
        self._frame_to_index: dict[SpeedscopeFrame, int] = {}

    def render_frame(self, frame: Frame | None) -> list[SpeedscopeEvent]:
        """
        Builds up a list of speedscope events that are used to populate the
        "events" array in speedscope-formatted JSON.

        This method has two notable side effects:

        * it populates the self._frame_to_index dictionary that matches
          speedscope frames with their positions in the "shared" array of
          speedscope output; this dictionary will be used to write this
          "shared" array in the render method

        * it accumulates a running total of time elapsed by
          accumulating the self_time spent in each pyinstrument frame;
          this running total is used by speedscope events to construct
          a flame chart.
        """

        # if frame is None, recursion bottoms out; no event frames
        # need to be added
        if frame is None:
            return []

        # Otherwise, form a speedscope frame and add it to the frame
        # to index map if the frame is not already a key in that map.
        sframe = SpeedscopeFrame(frame.function, frame.file_path, frame.line_no)
        if sframe not in self._frame_to_index:
            self._frame_to_index[sframe] = len(self._frame_to_index)

        # Get the frame index and add a speedscope event corresponding
        # to opening a stack frame.
        sframe_index = self._frame_to_index[sframe]
        open_event = SpeedscopeEvent(SpeedscopeEventType.OPEN, self._event_time, sframe_index)
        events_array: list[SpeedscopeEvent] = [open_event]

        # Add stack frame open and close events for all child frames
        # of this frame.
        for child in frame.children:
            events_array.extend(self.render_frame(child))

        # Update event time for closing this stack frame.
        #
        # If number of frames approaches 1e16 * desired accuracy
        # level, consider using Neumaier-Kahan summation; improves
        # worst-case relative accuracy of sum from O(num_summands *
        # eps) to (2 * eps + O(num_summands * eps * eps)), where eps
        # is IEEE-754 double precision unit roundoff, approximately
        # 1e-16. Average case relative accuracy expressions replace
        # num_summands with sqrt(num_summands). However, Kahan
        # summation quadruples operation count of sum, and Neumaier
        # variant also adds a branch & swap for each summand. Pairwise
        # summation isn't an option here because a running total is
        # needed.

        self._event_time += frame.absorbed_time
        if frame.is_synthetic_leaf:
            # only time contained within leaf nodes is real time i.e. not the sum of children
            self._event_time += frame.time

        # Add event closing this stack frame.
        close_event = SpeedscopeEvent(SpeedscopeEventType.CLOSE, self._event_time, sframe_index)
        events_array.append(close_event)

        return events_array

    def render(self, session: Session):
        frame = self.preprocess(session.root_frame())

        id_: str = time.strftime("%Y-%m-%dT%H-%M-%S", time.localtime(session.start_time))
        name: str = f"CPU profile for '{session.target_description}' at {id_}"

        sprofile_list: list[SpeedscopeProfile] = [
            SpeedscopeProfile(name, self.render_frame(frame), session.duration)
        ]

        # Exploits Python 3.7+ dictionary property of iterating over
        # keys in insertion order to build the list of speedscope
        # frames.
        sframe_list: list[SpeedscopeFrame] = [sframe for sframe in iter(self._frame_to_index)]

        shared_dict = {"frames": sframe_list}
        speedscope_file = SpeedscopeFile(name, sprofile_list, shared_dict)

        return "%s\n" % json.dumps(speedscope_file, cls=SpeedscopeEncoder)

    def default_processors(self) -> ProcessorList:
        """
        Default Processors for speedscope renderer; note that
        processors.aggregate_repeated_calls is removed because
        speedscope is a timeline-based format.
        """
        return [
            processors.remove_importlib,
            processors.remove_tracebackhide,
            processors.merge_consecutive_self_time,
            processors.remove_irrelevant_nodes,
            processors.remove_unnecessary_self_time_nodes,
            processors.remove_first_pyinstrument_frames_processor,
        ]



import asyncio
import os
import sys
import time
from typing import Callable, Generator, Generic, Iterable, Iterator, NoReturn, Optional, TypeVar

from flaky import flaky

from pyinstrument import stack_sampler
from pyinstrument.frame import SYNTHETIC_LEAF_IDENTIFIERS, Frame
from pyinstrument.profiler import Profiler
from pyinstrument.session import Session

if "CI" in os.environ:
    # a decorator that allows some test flakiness in CI environments, presumably
    # due to contention. Useful for tests that rely on real time measurements.
    flaky_in_ci = flaky(max_runs=5, min_passes=1)
else:
    flaky_in_ci = lambda a: a


def assert_never(x: NoReturn) -> NoReturn:
    raise AssertionError(f"Invalid value: {x!r}")


def do_nothing():
    pass


def busy_wait(duration):
    end_time = time.time() + duration

    while time.time() < end_time:
        do_nothing()


def walk_frames(frame: Frame) -> Generator[Frame, None, None]:
    yield frame

    for f in frame.children:
        yield from walk_frames(f)


T = TypeVar("T")


def first(iterator: Iterator[T]) -> Optional[T]:
    try:
        return next(iterator)
    except StopIteration:
        return None


def calculate_frame_tree_times(frame: Frame):
    # assuming that the leaf nodes of a frame tree have correct, time values,
    # calculate the times of all nodes in the frame tree

    child_time_sum = 0.0

    for child in frame.children:
        if child.identifier not in SYNTHETIC_LEAF_IDENTIFIERS:
            calculate_frame_tree_times(child)

        child_time_sum += child.time

    frame.time = child_time_sum + frame.absorbed_time


BUSY_WAIT_SCRIPT = """
import time, sys

def do_nothing():
    pass

def busy_wait(duration):
    end_time = time.time() + duration

    while time.time() < end_time:
        do_nothing()

def main():
    print('sys.argv: ', sys.argv)
    busy_wait(0.1)


if __name__ == '__main__':
    main()
"""


def dummy_session() -> Session:
    return Session(
        frame_records=[],
        start_time=0,
        min_interval=0.1,
        max_interval=0.1,
        duration=0,
        sample_count=0,
        start_call_stack=[],
        target_description="dummy",
        cpu_time=0,
        sys_path=sys.path,
        sys_prefixes=Session.current_sys_prefixes(),
    )


def tidy_up_profiler_state_on_fail(func: Callable) -> Callable[[], None]:
    """
    Useful inside a test that's flaky in CI, where the check_sampler_state
    fixture only gets to run at the end of all flaky attempts.
    """
    # consider adding to the flasky_in_ci decorator if it's useful elsewhere

    def wrapped():
        try:
            func()
        except BaseException:
            sys.setprofile(None)
            stack_sampler.thread_locals.__dict__.clear()
            raise

    return wrapped



#!/usr/bin/env python3
import argparse
import subprocess
import sys
from pathlib import Path

ROOT_DIR = Path(__file__).parent.parent
OUTPUT_DIR = ROOT_DIR / "html_renderer" / "demo-data"


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("SCRIPT", help="The script to run to produce the sample", type=Path)
    args = parser.parse_args()
    script_file: Path = args.SCRIPT
    output_file = (OUTPUT_DIR / script_file.with_suffix("").name).with_suffix(".json")

    result = subprocess.run(
        [
            "pyinstrument",
            "-o",
            str(output_file),
            "-r",
            "pyinstrument.renderers.html.JSONForHTMLRenderer",
            script_file,
        ]
    )

    if result.returncode != 0:
        return result.returncode

    print(f"Sample JSON written to {output_file}")


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

import codecs
import json
import tempfile
import urllib.parse
import warnings
import webbrowser
from pathlib import Path

from pyinstrument.renderers.base import FrameRenderer, ProcessorList, Renderer
from pyinstrument.session import Session

# pyright: strict


class HTMLRenderer(Renderer):
    """
    Renders a rich, interactive web page, as a string of HTML.
    """

    output_file_extension = "html"

    def __init__(
        self,
        show_all: bool = False,
        timeline: bool = False,
    ):
        super().__init__()
        if show_all:
            warnings.warn(
                f"the show_all option is deprecated on the HTML renderer, and has no effect. Use the view options in the webpage instead.",
                DeprecationWarning,
                stacklevel=3,
            )
        if timeline:
            warnings.warn(
                f"timeline is deprecated on the HTML renderer, and has no effect. Use the timeline view in the webpage instead.",
                DeprecationWarning,
                stacklevel=3,
            )
        # this is an undocumented option for use by the ipython magic, might
        # be removed later
        self.preprocessors: ProcessorList = []

    def render(self, session: Session):
        json_renderer = JSONForHTMLRenderer()
        json_renderer.processors = self.preprocessors
        session_json = json_renderer.render(session)

        resources_dir = Path(__file__).parent / "html_resources"

        js_file = resources_dir / "app.js"
        css_file = resources_dir / "app.css"

        if not js_file.exists() or not css_file.exists():
            raise RuntimeError(
                "Could not find app.js / app.css. Perhaps you need to run bin/build_js_bundle.py?"
            )

        js = js_file.read_text(encoding="utf-8")
        css = css_file.read_text(encoding="utf-8")

        page = f"""<!DOCTYPE html>
            <html>
            <head>
                <meta charset="utf-8">
            </head>
            <body>
                <div id="app"></div>

                <script>{js}</script>
                <style>{css}</style>

                <script>
                    const sessionData = {session_json};
                    pyinstrumentHTMLRenderer.render(document.getElementById('app'), sessionData);
                </script>
            </body>
            </html>
        """

        return page

    def open_in_browser(self, session: Session, output_filename: str | None = None):
        """
        Open the rendered HTML in a webbrowser.

        If output_filename=None (the default), a tempfile is used.

        The filename of the HTML file is returned.

        """
        if output_filename is None:
            output_file = tempfile.NamedTemporaryFile(suffix=".html", delete=False)
            output_filename = output_file.name
            with codecs.getwriter("utf-8")(output_file) as f:
                f.write(self.render(session))
        else:
            with codecs.open(output_filename, "w", "utf-8") as f:
                f.write(self.render(session))

        url = urllib.parse.urlunparse(("file", "", output_filename, "", "", ""))
        webbrowser.open(url)
        return output_filename


class JSONForHTMLRenderer(FrameRenderer):
    """
    The HTML takes a special form of JSON-encoded session, which includes
    an unprocessed frame tree rather than a list of frame records. This
    reduces the amount of parsing code that must be included in the
    Typescript renderer.
    """

    output_file_extension = "json"

    def default_processors(self) -> ProcessorList:
        return []

    def render(self, session: Session) -> str:
        session_json = session.to_json(include_frame_records=False)
        session_json_str = json.dumps(session_json)
        root_frame = session.root_frame()
        root_frame = self.preprocess(root_frame)
        frame_tree_json_str = root_frame.to_json_str() if root_frame else "null"
        return '{"session": %s, "frame_tree": %s}' % (session_json_str, frame_tree_json_str)



from __future__ import annotations

import asyncio
import html
import threading
import urllib.parse
from ast import parse
from textwrap import dedent

import IPython
from IPython import get_ipython  # type: ignore
from IPython.core.magic import Magics, line_cell_magic, magics_class, no_var_expand
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring
from IPython.display import IFrame, display

from pyinstrument import Profiler, renderers
from pyinstrument.__main__ import compute_render_options
from pyinstrument.frame import Frame
from pyinstrument.frame_ops import delete_frame_from_tree
from pyinstrument.processors import ProcessorOptions
from pyinstrument.renderers.console import ConsoleRenderer
from pyinstrument.renderers.html import HTMLRenderer

_active_profiler = None

_ASYNCIO_HTML_WARNING = """
To enable asyncio mode, use <pre>%%pyinstrument --async_mode=enabled</pre><br>
Note that due to IPython limitations this will run in a separate thread!
""".strip()
_ASYNCIO_TEXT_WARNING = (
    _ASYNCIO_HTML_WARNING.replace("<pre>", "`").replace("</pre>", "`").replace("<br>", "\n")
)


def _get_active_profiler():
    """
    Allows the code inserted into the cell to access the pyinstrument Profiler
    instance, to start/stop it.
    """
    return _active_profiler


class InterruptSilently(Exception):
    """Exception used to interrupt execution without showing traceback"""


@magics_class
class PyinstrumentMagic(Magics):
    def __init__(self, shell):
        super().__init__(shell)
        self._transformer = None

    def recreate_transformer(self, target_description: str):
        if IPython.version_info < (8, 15):  # type: ignore
            from ._utils import PrePostAstTransformer

            # This will leak _get_active_profiler into the users space until we can magle it
            pre = parse(
                dedent(
                    f"""
                    from pyinstrument.magic.magic import _get_active_profiler
                    _get_active_profiler().start(target_description={target_description!r})
                    """
                )
            )
            post = parse("\n_get_active_profiler().stop()")
            self._transformer = PrePostAstTransformer(pre, post)
        else:
            from IPython.core.magics.ast_mod import ReplaceCodeTransformer  # type: ignore

            self._transformer = ReplaceCodeTransformer.from_string(
                dedent(
                    f"""
                    from pyinstrument.magic.magic import _get_active_profiler as ___get_prof
                    ___get_prof().start(target_description={target_description!r})
                    try:
                        __code__
                    finally:
                        ___get_prof().stop()
                    __ret__
                    """
                )
            )

    @magic_arguments()
    @argument(
        "-p",
        "--render-option",
        dest="render_options",
        action="append",
        metavar="RENDER_OPTION",
        type=str,
        help=(
            "options to pass to the renderer, in the format 'flag_name' or 'option_name=option_value'. "
            "For example, to set the option 'time', pass '-p time=percent_of_total'. To pass multiple "
            "options, use the -p option multiple times. You can set processor options using dot-syntax, "
            "like '-p processor_options.filter_threshold=0'. option_value is parsed as a JSON value or "
            "a string."
        ),
    )
    @argument(
        "--show-regex",
        dest="show_regex",
        action="store",
        metavar="REGEX",
        help=(
            "regex matching the file paths whose frames to always show. "
            "Useful if --show doesn't give enough control."
        ),
    )
    @argument(
        "--show",
        dest="show_fnmatch",
        action="store",
        metavar="EXPR",
        help=(
            "glob-style pattern matching the file paths whose frames to "
            "show, regardless of --hide or --hide-regex. For example, use "
            "--show '*/<library>/*' to show frames within a library that "
            "would otherwise be hidden."
        ),
    )
    @argument(
        "--interval",
        type=float,
        default=0.001,
        help="The minimum time, in seconds, between each stack sample. See: https://pyinstrument.readthedocs.io/en/latest/reference.html#pyinstrument.Profiler.interval",
    )
    @argument(
        "--show-all",
        action="store_true",
        help="SHow all frames, including root frames with no time, and Internal IPython frames.",
    )
    @argument(
        "--async_mode",
        default="disabled",
        help="Configures how this Profiler tracks time in a program that uses async/await. See: https://pyinstrument.readthedocs.io/en/latest/reference.html#pyinstrument.Profiler.async_mode",
    )
    @argument(
        "--height",
        "-h",
        default=400,
        help="Output height",
    )
    @argument(
        "--timeline",
        type=bool,
        default=False,
        help="Show output timeline view",
    )
    @argument(
        "code",
        type=str,
        nargs="*",
        help="When used as a line magic, the code to profile",
    )
    @argument(
        "--hide",
        dest="hide_fnmatch",
        action="store",
        metavar="EXPR",
        help=(
            "glob-style pattern matching the file paths whose frames to hide. Defaults to "
            "hiding non-application code"
        ),
    )
    @argument(
        "--hide-regex",
        dest="hide_regex",
        action="store",
        metavar="REGEX",
        help=(
            "regex matching the file paths whose frames to hide. Useful if --hide doesn't give "
            "enough control."
        ),
    )
    @no_var_expand
    @line_cell_magic
    def pyinstrument(self, line, cell=None):
        """
        Run a cell with the pyinstrument statistical profiler.

        Converts the line/cell's AST to something like:
            try:
                profiler.start()
                run_code
            finally:
                profiler.stop()
            profiler.output_html()
        """
        global _active_profiler
        args = parse_argstring(self.pyinstrument, line)

        # 2024, always override this  for now in IPython,
        # we can make an option later if necessary
        args.unicode = True
        args.color = True

        ip = get_ipython()

        if not ip:
            raise RuntimeError("couldn't get ipython shell instance")

        if cell:
            target_description = f"Cell [{ip.execution_count}]"
        else:
            target_description = f"Line in cell [{ip.execution_count}]"
        code = cell or line

        if not code:
            return

        # Turn off the last run (e.g. a user interrupted)
        if _active_profiler and _active_profiler.is_running:
            _active_profiler.stop()
        if self._transformer in ip.ast_transformers:
            ip.ast_transformers.remove(self._transformer)

        _active_profiler = Profiler(interval=args.interval, async_mode=args.async_mode)
        self.recreate_transformer(target_description=target_description)
        ip.ast_transformers.append(self._transformer)
        if args.async_mode == "disabled":
            cell_result = ip.run_cell(code)
        else:
            cell_result = self.run_cell_async(ip, code)
        mangled_keys = [k for k in ip.user_ns.keys() if "-" in k]
        for k in mangled_keys:
            del ip.user_ns[k]
        ip.ast_transformers.remove(self._transformer)

        if (
            args.async_mode == "disabled"
            and cell_result.error_in_exec
            and isinstance(cell_result.error_in_exec, RuntimeError)
            and "event loop is already running" in str(cell_result.error_in_exec)
        ):
            # if the cell is async, the Magic doesn't work, raising the above
            # exception instead. We display a warning and return.
            display(
                {
                    "text/plain": _ASYNCIO_TEXT_WARNING,
                    "text/html": _ASYNCIO_HTML_WARNING,
                },
                raw=True,
            )
            return

        # If a KeyboardInterrupt occurred during the magic execution,
        # raise an exception to prevent further executions.
        if isinstance(cell_result.error_in_exec, KeyboardInterrupt):
            # The traceback is already shown during the cell execution above, so we
            # don't re-raise the exception directly.
            old_custom_tb = ip.CustomTB
            old_custom_exceptions = ip.custom_exceptions

            def _silent_exception_handler(self, etype, value, tb, tb_offset=None):
                # restore the original handlers
                ip.CustomTB = old_custom_tb
                ip.custom_exceptions = old_custom_exceptions
                # swallow the InterruptSilently entirely

            # install our silent handler
            ip.set_custom_exc((InterruptSilently,), _silent_exception_handler)
            raise InterruptSilently()

        html_config = compute_render_options(
            args, renderer_class=HTMLRenderer, unicode_support=True, color_support=True
        )

        text_config = compute_render_options(
            args, renderer_class=HTMLRenderer, unicode_support=True, color_support=True
        )

        html_renderer = renderers.HTMLRenderer(show_all=args.show_all, timeline=args.timeline)
        html_renderer.preprocessors.append(strip_ipython_frames_processor)
        html_str = _active_profiler.output(html_renderer)
        as_iframe = IFrame(
            src="data:text/html, Loading…",
            width="100%",
            height=args.height,
            extras=['style="resize: vertical"', f'srcdoc="{html.escape(html_str)}"'],
        )

        text_renderer = renderers.ConsoleRenderer(**text_config)
        text_renderer.processors.append(strip_ipython_frames_processor)

        as_text = _active_profiler.output(text_renderer)
        # repr_html may be a bit fragile, but it's been stable for a while
        display({"text/html": as_iframe._repr_html_(), "text/plain": as_text}, raw=True)  # type: ignore

        assert not _active_profiler.is_running
        _active_profiler = None

    def run_cell_async(self, ip, code):
        # This is a bit of a hack, but it's the only way to get the cell to run
        # asynchronously. We need to run the cell in a separate thread, and then
        # wait for it to finish.
        #
        # Please keep an eye on this issue to see if there's a better way:
        # https://github.com/ipython/ipython/issues/11314
        old_loop = asyncio.get_event_loop()
        loop = asyncio.new_event_loop()
        try:
            threading.Thread(target=loop.run_forever).start()
            asyncio.set_event_loop(loop)
            coro = ip.run_cell_async(code)
            future = asyncio.run_coroutine_threadsafe(coro, loop)
            return future.result()
        finally:
            loop.call_soon_threadsafe(loop.stop)
            asyncio.set_event_loop(old_loop)


IPYTHON_INTERNAL_FILES = (
    "IPython/core/interactiveshell.py",
    "IPython/terminal/interactiveshell.py",
    "IPython/core/async_helpers.py",
    "IPython/terminal/ipapp.py",
    "traitlets/config/application.py",
    "ipython/IPython/__init__.py",
    "ipykernel/zmqshell",
    "pyinstrument/magic/magic.py",
)


def strip_ipython_frames_processor(frame: Frame | None, options: ProcessorOptions) -> Frame | None:
    """
    A processor function that removes internal IPython nodes.
    """
    if frame is None:
        return None

    for child in frame.children:
        strip_ipython_frames_processor(child, options=options)

        if child.file_path is not None and any(
            f in child.file_path for f in IPYTHON_INTERNAL_FILES
        ):
            delete_frame_from_tree(child, replace_with="children")
            break

    return frame



from __future__ import annotations

import functools
import inspect
import sys
import typing

from pyinstrument.profiler import AsyncMode, Profiler
from pyinstrument.renderers.base import Renderer
from pyinstrument.renderers.console import ConsoleRenderer
from pyinstrument.typing import Unpack
from pyinstrument.util import file_supports_color, file_supports_unicode

CallableVar = typing.TypeVar("CallableVar", bound=typing.Callable)


class ProfileContextOptions(typing.TypedDict, total=False):
    interval: float
    async_mode: AsyncMode
    use_timing_thread: bool | None
    renderer: Renderer | None
    target_description: str | None


class ProfileContext:
    options: ProfileContextOptions

    def __init__(
        self,
        **kwargs: Unpack[ProfileContextOptions],
    ):
        profiler_options = {
            "interval": kwargs.get("interval", 0.001),
            # note- different async mode from the default, because it's easy
            # to run multiple profilers at once using the decorator/context
            # manager
            "async_mode": kwargs.get("async_mode", "disabled"),
            "use_timing_thread": kwargs.get("use_timing_thread", None),
        }
        self.profiler = Profiler(**profiler_options)
        self.options = kwargs

    @typing.overload
    def __call__(self, func: CallableVar, /) -> CallableVar: ...
    @typing.overload
    def __call__(self, /, **kwargs: Unpack[ProfileContextOptions]) -> "ProfileContext": ...
    def __call__(
        self, func: typing.Callable | None = None, /, **kwargs: Unpack[ProfileContextOptions]
    ):
        if func is not None:

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                target_description = self.options.get("target_description")
                if target_description is None:
                    target_description = f"Function {func.__qualname__} at {func.__code__.co_filename}:{func.__code__.co_firstlineno}"

                with self(target_description=target_description):
                    return func(*args, **kwargs)

            return typing.cast(typing.Callable, wrapper)
        else:
            return ProfileContext(**{**self.options, **kwargs})

    def __enter__(self):
        if self.profiler.is_running:
            raise RuntimeError(
                "This profiler is already running - did you forget the brackets on pyinstrument.profile() ?"
            )

        caller_frame = inspect.currentframe().f_back  # type: ignore
        assert caller_frame is not None
        target_description = self.options.get("target_description")
        if target_description is None:
            target_description = "Block at {}:{}".format(
                caller_frame.f_code.co_filename, caller_frame.f_lineno
            )

        self.profiler.start(
            caller_frame=caller_frame,
            target_description=target_description,
        )

    def __exit__(self, exc_type, exc_value, traceback):
        session = self.profiler.stop()

        renderer = self.options.get("renderer")
        f = sys.stderr

        if renderer is None:
            renderer = ConsoleRenderer(
                color=file_supports_color(f),
                unicode=file_supports_unicode(f),
                short_mode=True,
            )

        f.write(renderer.render(session))


class _Profile:
    @typing.overload
    def __call__(self, func: CallableVar, /) -> CallableVar: ...
    @typing.overload
    def __call__(self, /, **kwargs: Unpack[ProfileContextOptions]) -> "ProfileContext": ...
    def __call__(
        self, func: typing.Callable | None = None, /, **kwargs: Unpack[ProfileContextOptions]
    ):
        if func is not None:
            return ProfileContext(**kwargs)(func)
        else:
            return ProfileContext(**kwargs)


profile = _Profile()



import signal
import textwrap
from test.fake_time_util import fake_time
from threading import Thread
from time import sleep

import pytest

# note: IPython should be imported within each test. Importing it in our tests
# seems to cause problems with subsequent tests.

cell_code = """
import time

def function_a():
    function_b()
    function_c()

def function_b():
    function_d()

def function_c():
    function_d()

def function_d():
    function_e()

def function_e():
    time.sleep(0.1)

function_a()
"""

# Tests #


@pytest.mark.ipythonmagic
def test_magics(ip):
    from IPython.utils.capture import capture_output as capture_ipython_output

    with fake_time():
        with capture_ipython_output() as captured:
            ip.run_cell_magic("pyinstrument", line="", cell=cell_code)

    assert len(captured.outputs) == 1
    output = captured.outputs[0]
    assert "text/html" in output.data
    assert "text/plain" in output.data

    assert "function_a" in output.data["text/html"]
    assert "<iframe" in output.data["text/html"]
    assert "function_a" in output.data["text/plain"]

    assert "- 0.200 function_a" in output.data["text/plain"]
    assert "- 0.100 FakeClock.sleep" in output.data["text/plain"]

    with fake_time():
        with capture_ipython_output() as captured:
            # this works because function_a was defined in the previous cell
            ip.run_line_magic("pyinstrument", line="function_a()")

    assert len(captured.outputs) == 1
    output = captured.outputs[0]

    assert "function_a" in output.data["text/plain"]
    assert "- 0.100 FakeClock.sleep" in output.data["text/plain"]


@pytest.mark.ipythonmagic
def test_magic_empty_line(ip):
    # check empty line input
    ip.run_line_magic("pyinstrument", line="")


@pytest.mark.ipythonmagic
def test_magic_no_variable_expansion(ip, capsys):
    ip.run_line_magic("pyinstrument", line="print(\"hello {len('world')}\")")

    captured = capsys.readouterr()
    assert "hello {len('world')}" in captured.out
    assert "hello 5" not in captured.out


@pytest.mark.ipythonmagic
def test_pyinstrument_handles_interrupt_silently(ip, capsys):
    from pyinstrument.magic.magic import InterruptSilently

    thread = Thread(target=_interrupt_after_1s)
    thread.start()
    # expect our custom exception to bubble up
    with pytest.raises(InterruptSilently):
        ip.run_cell_magic("pyinstrument", "", "from time import sleep; sleep(2)")

    thread.join()
    # nothing should have hit stderr
    _, err = capsys.readouterr()
    assert err.strip() == ""


@pytest.mark.ipythonmagic
def test_async_cell_with_pyinstrument(ip, capsys):
    ip.run_cell_magic(
        "pyinstrument",
        line="--async_mode=enabled",
        cell=textwrap.dedent(
            """
            import asyncio
            async def function_a():
                await asyncio.sleep(0.1)
                return 42
            a = await function_a()
            print("a:", a)
            """
        ),
    )
    stdout, stderr = capsys.readouterr()
    assert "a: 42" in stdout


# Utils #


@pytest.fixture(scope="module")
def session_ip():
    from IPython.testing.globalipapp import start_ipython

    yield start_ipython()


def _interrupt_after_1s():
    sleep(1)
    signal.raise_signal(signal.SIGINT)


@pytest.fixture(scope="function")
def ip(session_ip):
    session_ip.run_line_magic(magic_name="load_ext", line="pyinstrument")
    yield session_ip
    session_ip.run_line_magic(magic_name="reset", line="-f")



import os
from typing import TYPE_CHECKING, Any, Union

if TYPE_CHECKING:
    from typing_extensions import Literal as LiteralStr
    from typing_extensions import TypeAlias, Unpack, assert_never
else:
    # a type, that when subscripted, returns `str`.
    class _LiteralStr:
        def __getitem__(self, values):
            return str

    LiteralStr = _LiteralStr()

    def assert_never(value: Any):
        raise ValueError(value)

    Unpack = Any
    TypeAlias = Any


PathOrStr = Union[str, "os.PathLike[str]"]

__all__ = ["PathOrStr", "LiteralStr", "assert_never", "Unpack", "TypeAlias"]



# keypath vendored from https://github.com/fictorial/keypath

# keypath is released under the BSD license:

# Copyright 2016, Fictorial LLC

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

#     * Redistributions of source code must retain the above
#       copyright notice, this list of conditions and the following
#       disclaimer.

#     * Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials
#       provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER “AS IS” AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# Includes modifications by joerick, which fall under the same license.

from typing import Any


def value_at_keypath(obj: Any, keypath: str) -> Any:
  """
  Returns value at given key path which follows dotted-path notation.

    >>> x = dict(a=1, b=2, c=dict(d=3, e=4, f=[2,dict(x='foo', y='bar'),5]))
    >>> assert value_at_keypath(x, 'a') == 1
    >>> assert value_at_keypath(x, 'b') == 2
    >>> assert value_at_keypath(x, 'c.d') == 3
    >>> assert value_at_keypath(x, 'c.e') == 4
    >>> assert value_at_keypath(x, 'c.f.0') == 2
    >>> assert value_at_keypath(x, 'c.f.-1') == 5
    >>> assert value_at_keypath(x, 'c.f.1.y') == 'bar'

  """
  for part in keypath.split('.'):
    if isinstance(obj, dict):
      obj = obj.get(part, {})
    elif type(obj) in [tuple, list]:
      obj = obj[int(part)]
    else:
      obj = getattr(obj, part, {})
  return obj


def set_value_at_keypath(obj: Any, keypath: str, val: Any):
  """
  Sets value at given key path which follows dotted-path notation.

  Each part of the keypath must already exist in the target value
  along the path.

    >>> x = dict(a=1, b=2, c=dict(d=3, e=4, f=[2,dict(x='foo', y='bar'),5]))
    >>> assert set_value_at_keypath(x, 'a', 2)
    >>> assert value_at_keypath(x, 'a') == 2
    >>> assert set_value_at_keypath(x, 'c.f.-1', 6)
    >>> assert value_at_keypath(x, 'c.f.-1') == 6
  """
  parts = keypath.split('.')
  for part in parts[:-1]:
    if isinstance(obj, dict):
      obj = obj[part]
    elif type(obj) in [tuple, list]:
      obj = obj[int(part)]
    else:
      obj = getattr(obj, part)
  last_part = parts[-1]
  if isinstance(obj, dict):
    obj[last_part] = val
  elif type(obj) in [tuple, list]:
    obj[int(last_part)] = val
  else:
    setattr(obj, last_part, val)
  return True



from __future__ import annotations

import threading
import time
from typing import Any, List
from unittest import TestCase

import pytest

from pyinstrument.low_level.stat_profile import setstatprofile

from ..util import busy_wait, do_nothing


class CallCounter:
    def __init__(self, thread) -> None:
        self.thread = thread
        self.count = 0

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        assert self.thread is threading.current_thread()
        self.count += 1


def test_threaded():
    # assert that each thread gets its own callbacks, and check that it
    # doesn't crash!

    counters: list[CallCounter | None] = [None for _ in range(10)]
    stop = False

    def profile_a_busy_wait(i):
        thread = threads[i]
        counter = CallCounter(thread)
        counters[i] = counter

        setstatprofile(counter, 0.001)
        while not stop:
            do_nothing()
        setstatprofile(None)

    threads = [threading.Thread(target=profile_a_busy_wait, args=(i,)) for i in range(10)]
    for thread in threads:
        thread.start()

    while not stop:
        stop = all(c is not None and c.count > 10 for c in counters)

    for thread in threads:
        thread.join()



import contextvars
import sys
import time

import pytest

from pyinstrument import stack_sampler

from .util import do_nothing, flaky_in_ci, tidy_up_profiler_state_on_fail


class SampleCounter:
    count = 0

    def sample(self, stack, time, async_state):
        self.count += 1


def test_create():
    sampler = stack_sampler.get_stack_sampler()
    assert sampler is not None

    assert sampler is stack_sampler.get_stack_sampler()


@flaky_in_ci
@tidy_up_profiler_state_on_fail
def test_get_samples():
    sampler = stack_sampler.get_stack_sampler()
    counter = SampleCounter()

    assert sys.getprofile() is None
    sampler.subscribe(counter.sample, desired_interval=0.001, use_async_context=True)
    assert sys.getprofile() is not None
    assert len(sampler.subscribers) == 1

    start = time.time()
    while time.time() < start + 1 and counter.count == 0:
        do_nothing()

    assert counter.count > 0

    assert sys.getprofile() is not None
    sampler.unsubscribe(counter.sample)
    assert sys.getprofile() is None

    assert len(sampler.subscribers) == 0


@flaky_in_ci
@tidy_up_profiler_state_on_fail
def test_multiple_samplers():
    sampler = stack_sampler.get_stack_sampler()
    counter_1 = SampleCounter()
    counter_2 = SampleCounter()

    sampler.subscribe(counter_1.sample, desired_interval=0.001, use_async_context=False)
    sampler.subscribe(counter_2.sample, desired_interval=0.001, use_async_context=False)

    assert len(sampler.subscribers) == 2

    start = time.time()
    while time.time() < start + 1 and counter_1.count == 0 and counter_2.count == 0:
        do_nothing()

    assert counter_1.count > 0
    assert counter_2.count > 0

    assert sys.getprofile() is not None

    sampler.unsubscribe(counter_1.sample)
    sampler.unsubscribe(counter_2.sample)

    assert sys.getprofile() is None

    assert len(sampler.subscribers) == 0


def test_multiple_samplers_async_error():
    sampler = stack_sampler.get_stack_sampler()

    counter_1 = SampleCounter()
    counter_2 = SampleCounter()

    sampler.subscribe(counter_1.sample, desired_interval=0.001, use_async_context=True)

    with pytest.raises(RuntimeError):
        sampler.subscribe(counter_2.sample, desired_interval=0.001, use_async_context=True)

    sampler.unsubscribe(counter_1.sample)


@flaky_in_ci
@tidy_up_profiler_state_on_fail
def test_multiple_contexts():
    sampler = stack_sampler.get_stack_sampler()

    counter_1 = SampleCounter()
    counter_2 = SampleCounter()

    context_1 = contextvars.copy_context()
    context_2 = contextvars.copy_context()

    assert sys.getprofile() is None
    assert len(sampler.subscribers) == 0
    context_1.run(
        sampler.subscribe, target=counter_1.sample, desired_interval=0.001, use_async_context=True
    )
    context_2.run(
        sampler.subscribe, target=counter_2.sample, desired_interval=0.001, use_async_context=True
    )

    assert sys.getprofile() is not None
    assert len(sampler.subscribers) == 2

    start = time.time()
    while time.time() < start + 1 and counter_1.count == 0 and counter_2.count == 0:
        do_nothing()

    assert counter_1.count > 0
    assert counter_2.count > 0

    assert sys.getprofile() is not None

    context_1.run(sampler.unsubscribe, counter_1.sample)
    context_2.run(sampler.unsubscribe, counter_2.sample)

    assert sys.getprofile() is None

    assert len(sampler.subscribers) == 0



import os
import time
from pathlib import Path
from pstats import Stats
from test.fake_time_util import FakeClock, fake_time
from typing import Any

import pytest

from pyinstrument import Profiler
from pyinstrument.renderers import PstatsRenderer


def a():
    b()
    c()


def b():
    d()


def c():
    d()


def d():
    e()


def e():
    time.sleep(1)


@pytest.fixture(scope="module")
def profiler_session():
    with fake_time():
        profiler = Profiler()
        profiler.start()

        a()

        profiler.stop()
        return profiler.last_session


def test_pstats_renderer(profiler_session, tmp_path):
    fname = tmp_path / "test.pstats"
    pstats_data = PstatsRenderer().render(profiler_session)
    with open(fname, "wb") as fid:
        fid.write(pstats_data.encode(encoding="utf-8", errors="surrogateescape"))
    stats: Any = Stats(str(fname))
    # Sanity check
    assert stats.total_tt > 0
    # The graph is
    # a() -> b() -> d() -> e() -> time.sleep()
    #    \-> c() /
    # so make sure d has callers of b, c, and that the times make sense

    # in stats,
    #   keys are tuples (file_path, line, func)
    #   values are tuples (calltime, numcalls, selftime, cumtime, callers)
    # in callers,
    #   keys are the same as in stats
    #   values are the same as stats but without callers

    # check the time of d
    d_key = [k for k in stats.stats.keys() if k[2] == "d"][0]
    d_val = stats.stats[d_key]
    d_cumtime = d_val[3]
    assert d_cumtime == pytest.approx(2)

    # check d's callers times are split
    b_key = [k for k in stats.stats.keys() if k[2] == "b"][0]
    c_key = [k for k in stats.stats.keys() if k[2] == "c"][0]
    d_callers = d_val[4]
    b_cumtime = d_callers[b_key][3]
    c_cumtime = d_callers[c_key][3]
    assert b_cumtime == pytest.approx(1)
    assert c_cumtime == pytest.approx(1)

    # check the time of e
    e_key = [k for k in stats.stats.keys() if k[2] == "e"][0]
    e_val = stats.stats[e_key]
    e_cumtime = e_val[3]
    assert e_cumtime == pytest.approx(2)


def test_round_trip_encoding_of_binary_data(tmp_path: Path):
    # as used by the pstats renderer
    data_blob = os.urandom(1024)
    file = tmp_path / "file.dat"

    data_blob_string = data_blob.decode(encoding="utf-8", errors="surrogateescape")

    # newline='' is required to prevent the default newline translation
    with open(file, mode="w", encoding="utf-8", errors="surrogateescape", newline="") as f:
        f.write(data_blob_string)

    assert data_blob == data_blob_string.encode(encoding="utf-8", errors="surrogateescape")
    assert data_blob == file.read_bytes()


def sleep_and_busy_wait(clock: FakeClock):
    time.sleep(1.0)
    # this looks like a busy wait to the profiler
    clock.time += 1.0


def test_sum_of_tottime(tmp_path):
    # Check that the sum of the tottime of all the functions is equal to the
    # total time of the profile

    with fake_time() as clock:
        profiler = Profiler()
        profiler.start()

        sleep_and_busy_wait(clock)

        profiler.stop()
        profiler_session = profiler.last_session

    assert profiler_session

    pstats_data = PstatsRenderer().render(profiler_session)
    fname = tmp_path / "test.pstats"
    with open(fname, "wb") as fid:
        fid.write(pstats_data.encode(encoding="utf-8", errors="surrogateescape"))
    stats: Any = Stats(str(fname))
    assert stats.total_tt == pytest.approx(2)



from __future__ import annotations

import contextvars
import sys
import timeit
import types
from typing import Any, Callable, List, Optional, Type

from pyinstrument.low_level.pyi_timing_thread_python import (
    pyi_timing_thread_get_time,
    pyi_timing_thread_subscribe,
    pyi_timing_thread_unsubscribe,
)
from pyinstrument.low_level.types import TimerType


class PythonStatProfiler:
    await_stack: list[str]
    timing_thread_subscription: int | None = None

    def __init__(
        self,
        target: Callable[[types.FrameType, str, Any], Any],
        interval: float,
        context_var: contextvars.ContextVar[object | None] | None,
        timer_type: TimerType,
        timer_func: Callable[[], float] | None,
    ):
        self.target = target
        self.interval = interval
        if context_var:
            # raise typeerror to match the C version
            if not isinstance(context_var, contextvars.ContextVar):
                raise TypeError("not a context var")
        self.context_var = context_var

        self.timer_type = timer_type

        if timer_type == "walltime":
            self.get_time = timeit.default_timer
        elif timer_type == "walltime_thread":
            self.get_time = pyi_timing_thread_get_time
            self.timing_thread_subscription = pyi_timing_thread_subscribe(interval)
        elif timer_type == "timer_func":
            if timer_func is None:
                raise TypeError("timer_func must be provided for timer_func timer_type")
            self.get_time = timer_func
        else:
            raise ValueError(f"invalid timer_type '{timer_type}'")

        self.last_invocation = self.get_time()

        self.last_context_var_value = context_var.get() if context_var else None
        self.await_stack = []

    def __del__(self):
        if self.timing_thread_subscription is not None:
            pyi_timing_thread_unsubscribe(self.timing_thread_subscription)

    def profile(self, frame: types.FrameType, event: str, arg: Any):
        now = self.get_time()

        if self.context_var:
            context_var_value = self.context_var.get()
            last_context_var_value = self.last_context_var_value

            if context_var_value is not last_context_var_value:
                context_change_frame = frame.f_back if event == "call" else frame
                assert context_change_frame is not None
                self.target(
                    context_change_frame,
                    "context_changed",
                    (context_var_value, last_context_var_value, self.await_stack),
                )
                self.last_context_var_value = context_var_value

            # 0x80 == CO_COROUTINE (i.e. defined with 'async def')
            if event == "return" and frame.f_code.co_flags & 0x80:
                self.await_stack.append(get_frame_info(frame))
            else:
                self.await_stack.clear()

        if now < self.last_invocation + self.interval:
            return

        self.last_invocation = now
        return self.target(frame, event, arg)


"""
A reimplementation of setstatprofile in Python, for prototyping/reference
purposes. Not used in normal execution.
"""


def setstatprofile(
    target: Callable[[types.FrameType, str, Any], Any] | None,
    interval: float = 0.001,
    context_var: contextvars.ContextVar[object | None] | None = None,
    timer_type: TimerType = "walltime",
    timer_func: Callable[[], float] | None = None,
) -> None:
    if target:
        profiler = PythonStatProfiler(
            target=target,
            interval=interval,
            context_var=context_var,
            timer_type=timer_type,
            timer_func=timer_func,
        )
        sys.setprofile(profiler.profile)
    else:
        sys.setprofile(None)


def get_frame_info(frame: types.FrameType) -> str:
    frame_info = "%s\x00%s\x00%i" % (
        frame.f_code.co_name,
        frame.f_code.co_filename,
        frame.f_code.co_firstlineno,
    )

    class_name = None
    # try to find self argument for methods
    self = frame.f_locals.get("self", None)
    if self and hasattr(self, "__class__") and hasattr(self.__class__, "__qualname__"):
        class_name = self.__class__.__qualname__
    else:
        # also try to find cls argument for class methods
        cls = frame.f_locals.get("cls", None)
        if cls and hasattr(cls, "__qualname__"):
            class_name = cls.__qualname__

    frame_hidden = "__tracebackhide__" in frame.f_locals

    if class_name:
        frame_info += "\x01c%s" % class_name

    if frame.f_lineno is not None:
        frame_info += "\x01l%i" % frame.f_lineno

    if frame_hidden:
        frame_info += "\x01h%i" % frame_hidden

    return frame_info



from pyinstrument.renderers.base import FrameRenderer, Renderer
from pyinstrument.renderers.console import ConsoleRenderer
from pyinstrument.renderers.html import HTMLRenderer
from pyinstrument.renderers.jsonrenderer import JSONRenderer
from pyinstrument.renderers.pstatsrenderer import PstatsRenderer
from pyinstrument.renderers.session import SessionRenderer
from pyinstrument.renderers.speedscope import SpeedscopeRenderer

__all__ = [
    "ConsoleRenderer",
    "FrameRenderer",
    "HTMLRenderer",
    "JSONRenderer",
    "PstatsRenderer",
    "Renderer",
    "SessionRenderer",
    "SpeedscopeRenderer",
]



import warnings

from pyinstrument.context_manager import profile
from pyinstrument.profiler import Profiler

__all__ = ["__version__", "Profiler", "load_ipython_extension", "profile"]
__version__ = "5.0.3"

# enable deprecation warnings
warnings.filterwarnings("once", ".*", DeprecationWarning, r"pyinstrument\..*")


def load_ipython_extension(ipython):
    """
    This function is called by IPython to load the pyinstrument IPython
    extension, which is done with the magic command `%load_ext pyinstrument`.
    """

    from pyinstrument.magic import PyinstrumentMagic

    ipython.register_magics(PyinstrumentMagic)



import ctypes
import time

import pytest

import pyinstrument.low_level.stat_profile as native_module

lib = ctypes.CDLL(native_module.__file__)

pyi_floatclock = lib.pyi_floatclock
pyi_floatclock.argtypes = [ctypes.c_int]
pyi_floatclock.restype = ctypes.c_double


def test_floatclock():
    time_a = pyi_floatclock(0)
    time.sleep(0.001)
    time_b = pyi_floatclock(0)
    assert time_b > time_a


def test_is_in_seconds():
    floatclock_time_a = pyi_floatclock(0)
    time_a = time.time()

    time.sleep(0.1)

    floatclock_time_b = pyi_floatclock(0)
    time_b = time.time()

    floatclock_duration = floatclock_time_b - floatclock_time_a
    duration = time_b - time_a

    assert floatclock_duration == pytest.approx(duration, rel=0.1)



import ctypes
import os
import sys
import time

import pyinstrument.low_level.stat_profile as native_module

from ..util import busy_wait, flaky_in_ci

lib = ctypes.CDLL(native_module.__file__)

pyi_timing_thread_subscribe = lib.pyi_timing_thread_subscribe
pyi_timing_thread_subscribe.argtypes = [ctypes.c_double]
pyi_timing_thread_subscribe.restype = ctypes.c_int

pyi_timing_thread_get_time = lib.pyi_timing_thread_get_time
pyi_timing_thread_get_time.argtypes = []
pyi_timing_thread_get_time.restype = ctypes.c_double

pyi_timing_thread_get_interval = lib.pyi_timing_thread_get_interval
pyi_timing_thread_get_interval.argtypes = []
pyi_timing_thread_get_interval.restype = ctypes.c_double

pyi_timing_thread_unsubscribe = lib.pyi_timing_thread_unsubscribe
pyi_timing_thread_unsubscribe.argtypes = [ctypes.c_int]
pyi_timing_thread_unsubscribe.restype = ctypes.c_int

PYI_TIMING_THREAD_UNKNOWN_ERROR = -1
PYI_TIMING_THREAD_TOO_MANY_SUBSCRIBERS = -2


if sys.platform == "win32":
    # on windows, the thread scheduling 'quanta', the time that a thread can run
    # before potentially being pre-empted, is 20-30ms. This means that the
    # worst-case, we have to wait 30ms before the timing thread gets a chance to
    # run. This isn't really a huge problem in practice, because thread-based
    # timing isn't much use on windows, since the synchronous timing functions are
    # so fast.
    WAIT_TIME = 0.03
elif os.environ.get("QEMU_EMULATED"):
    # the scheduler seems slower under emulation
    WAIT_TIME = 0.2
else:
    WAIT_TIME = 0.015


@flaky_in_ci
def test():
    # check the thread isn't running to begin with
    assert pyi_timing_thread_get_interval() == -1

    time_before = pyi_timing_thread_get_time()
    time.sleep(WAIT_TIME)
    assert pyi_timing_thread_get_time() == time_before

    # subscribe
    subscription_id = pyi_timing_thread_subscribe(0.001)
    try:
        assert subscription_id >= 0

        assert pyi_timing_thread_get_interval() == 0.001

        # check it's updating
        busy_wait(WAIT_TIME)
        time_a = pyi_timing_thread_get_time()
        assert time_a > time_before
        busy_wait(WAIT_TIME)
        time_b = pyi_timing_thread_get_time()
        assert time_b > time_a

        # unsubscribe
        assert pyi_timing_thread_unsubscribe(subscription_id) == 0

        assert pyi_timing_thread_get_interval() == -1

        # check it's stopped updating
        time.sleep(WAIT_TIME)
        time_c = pyi_timing_thread_get_time()
        time.sleep(WAIT_TIME)
        time_d = pyi_timing_thread_get_time()
        assert time_c == time_d
    finally:
        # ensure the subscriber is removed even if the test fails
        pyi_timing_thread_unsubscribe(subscription_id)


def test_max_subscribers():
    subscription_ids = []

    try:
        for i in range(1000):
            subscription_id = pyi_timing_thread_subscribe(0.001)
            assert subscription_id >= 0
            subscription_ids.append(subscription_id)

        # the next one should fail
        assert pyi_timing_thread_subscribe(0.001) == PYI_TIMING_THREAD_TOO_MANY_SUBSCRIBERS

        # unsubscribe them in FIFO order
        for subscription_id in subscription_ids:
            assert pyi_timing_thread_get_interval() == 0.001
            assert pyi_timing_thread_unsubscribe(subscription_id) == 0

        # check there are no subscribers left
        assert pyi_timing_thread_get_interval() == -1
    finally:
        # ensure all subscription ids are removed even if the test fails
        while subscription_ids:
            subscription_id = subscription_ids.pop()
            pyi_timing_thread_unsubscribe(subscription_id)



import codecs
import importlib
import math
import os
import re
import sys
import warnings
from typing import IO, Any, AnyStr, Callable

from pyinstrument.vendor.decorator import decorator


def object_with_import_path(import_path: str) -> Any:
    if "." not in import_path:
        raise ValueError("Can't import '%s', it is not a valid import path" % import_path)
    module_path, object_name = import_path.rsplit(".", 1)

    module = importlib.import_module(module_path)
    return getattr(module, object_name)


def truncate(string: str, max_length: int) -> str:
    if len(string) > max_length:
        return string[0 : max_length - 3] + "..."
    return string


@decorator
def deprecated(func: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
    """Marks a function as deprecated."""
    warnings.warn(
        f"{func} is deprecated and should no longer be used.",
        DeprecationWarning,
        stacklevel=3,
    )
    return func(*args, **kwargs)


def deprecated_option(option_name: str, message: str = "") -> Any:
    """Marks an option as deprecated."""

    def caller(func, *args, **kwargs):
        if option_name in kwargs:
            warnings.warn(
                f"{option_name} is deprecated. {message}",
                DeprecationWarning,
                stacklevel=3,
            )

        return func(*args, **kwargs)

    return decorator(caller)


def file_supports_color(file_obj: IO[AnyStr]) -> bool:
    """
    Returns True if the running system's terminal supports color.

    Borrowed from Django
    https://github.com/django/django/blob/master/django/core/management/color.py
    """
    plat = sys.platform
    supported_platform = plat != "Pocket PC" and (plat != "win32" or "ANSICON" in os.environ)

    is_a_tty = file_is_a_tty(file_obj)

    return supported_platform and is_a_tty


def file_supports_unicode(file_obj: IO[AnyStr]) -> bool:
    encoding = getattr(file_obj, "encoding", None)
    if not encoding:
        return False

    codec_info = codecs.lookup(encoding)

    return "utf" in codec_info.name


def file_is_a_tty(file_obj: IO[AnyStr]) -> bool:
    return hasattr(file_obj, "isatty") and file_obj.isatty()


def unwrap(string: str) -> str:
    string = string.replace("\n", " ")
    string = re.sub(r"\s+", " ", string)
    return string.strip()


def format_float_with_sig_figs(value: float, sig_figs: int = 3, trim_zeroes=False) -> str:
    """
    Format a float to a string with a specific number of significant figures.
    Doesn't use scientific notation.
    """
    if value == 0:
        return "0"

    precision = math.ceil(-math.log10(abs(value))) + sig_figs - 1
    if precision < 0:
        precision = 0
    result = "{:.{precision}f}".format(value, precision=precision)

    if trim_zeroes and "." in result:
        result = result.rstrip("0").rstrip(".")

    return result


def strtobool(val: str) -> bool:
    return val.lower() in {"y", "yes", "t", "true", "on", "1"}



import json

from pyinstrument.renderers.base import Renderer
from pyinstrument.session import Session


class SessionRenderer(Renderer):
    output_file_extension: str = "pyisession"

    def __init__(self, tree_format: bool = False):
        super().__init__()
        self.tree_format = tree_format

    def render(self, session: Session) -> str:
        return json.dumps(session.to_json())



from __future__ import annotations

import re
import textwrap
import time
from typing import Any, Dict, List, Tuple

import pyinstrument
from pyinstrument import processors
from pyinstrument.frame import Frame, FrameGroup
from pyinstrument.renderers.base import FrameRenderer, ProcessorList, Renderer
from pyinstrument.session import Session
from pyinstrument.typing import LiteralStr
from pyinstrument.util import truncate

# pyright: strict

FlatTimeMode = LiteralStr["self", "total"]


class ConsoleRenderer(FrameRenderer):
    """
    Produces text-based output, suitable for text files or ANSI-compatible
    consoles.
    """

    def __init__(
        self,
        show_all: bool = False,
        timeline: bool = False,
        processor_options: dict[str, Any] | None = None,
        unicode: bool = False,
        color: bool = False,
        flat: bool = False,
        time: LiteralStr["seconds", "percent_of_total"] = "seconds",
        flat_time: FlatTimeMode = "self",
        short_mode: bool = False,
    ) -> None:
        """
        :param unicode: Use unicode, like box-drawing characters in the output.
        :param color: Enable color support, using ANSI color sequences.
        :param flat: Display a flat profile instead of a call graph.
        :param time: How to display the duration of each frame - ``'seconds'`` or ``'percent_of_total'``
        :param flat_time: Show ``'self'`` time or ``'total'`` time (including children) in flat profile.
        :param short_mode: Display a short version of the output.
        :param show_all: See :class:`FrameRenderer`.
        :param timeline: See :class:`FrameRenderer`.
        :param processor_options: See :class:`FrameRenderer`.
        """
        super().__init__(show_all=show_all, timeline=timeline, processor_options=processor_options)
        self.unicode = unicode
        self.color = color
        self.flat = flat
        self.time = time
        self.flat_time = flat_time
        self.short_mode = short_mode

        if self.flat and self.timeline:
            raise Renderer.MisconfigurationError("Cannot use timeline and flat options together.")

        self.colors = self.colors_enabled if color else self.colors_disabled

    def render(self, session: Session) -> str:
        result = self.render_preamble(session)

        frame = self.preprocess(session.root_frame())
        indent = ".  " if self.short_mode else ""

        if frame is None:
            result += f"{indent}No samples were recorded.\n"
        else:
            self.root_frame = frame

            if self.flat:
                result += self.render_frame_flat(self.root_frame, indent=indent)
            else:
                result += self.render_frame(self.root_frame, indent=indent, child_indent=indent)

        result += f"{indent}\n"

        if self.short_mode:
            result += "." * 53 + "\n\n"

        return result

    # pylint: disable=W1401
    def render_preamble(self, session: Session) -> str:
        if self.short_mode:
            return textwrap.dedent(
                f"""
                    pyinstrument ........................................
                    .
                    .  {session.target_description}
                    .
                """
            )

        lines = [
            r"",
            r"  _     ._   __/__   _ _  _  _ _/_  ",
            r" /_//_/// /_\ / //_// / //_'/ //    ",
            r"/   _/        {:>20}".format("v" + pyinstrument.__version__),
        ]

        lines[1] += " Recorded: {:<9}".format(
            time.strftime("%X", time.localtime(session.start_time))
        )
        lines[2] += f" Duration: {session.duration:<9.3f}"
        lines[1] += f" Samples:  {session.sample_count}"
        lines[2] += f" CPU time: {session.cpu_time:.3f}"

        lines.append("")
        lines.append(session.target_description)
        lines.append("")
        lines.append("")

        return "\n".join(lines)

    def should_render_frame(self, frame: Frame) -> bool:
        if frame.group and not self.should_ignore_group(frame.group):
            return self.should_render_frame_in_group(frame)
        return True

    def should_render_frame_in_group(self, frame: Frame) -> bool:
        # Only render the root frame, or frames that are significant
        assert frame.group
        return (
            frame.group.root == frame
            or frame.total_self_time > 0.2 * self.root_frame.time
            or frame in frame.group.exit_frames
        )

    def should_ignore_group(self, group: FrameGroup) -> bool:
        """
        If a group is ignored, its frames are all printed - they're not hidden.
        """
        hidden_frames = [f for f in group.frames if not self.should_render_frame_in_group(f)]
        # don't bother printing groups with one/zero hidden frames
        return len(hidden_frames) < 2

    def group_description(self, group: FrameGroup) -> str:
        hidden_frames = [f for f in group.frames if not self.should_render_frame(f)]
        libraries = self.libraries_for_frames(hidden_frames)
        return "[{count} frames hidden]  {c.faint}{libraries}{c.end}\n".format(
            count=len(hidden_frames),
            libraries=truncate(", ".join(libraries), 40),
            c=self.colors,
        )

    def libraries_for_frames(self, frames: list[Frame]) -> list[str]:
        libraries: list[str] = []
        for frame in frames:
            if frame.file_path_short:
                library = re.split(r"[\\/\.]", frame.file_path_short, maxsplit=1)[0]

                if library and library not in libraries:
                    libraries.append(library)
        return libraries

    def render_frame(self, frame: Frame, indent: str = "", child_indent: str = "") -> str:
        if self.should_render_frame(frame):
            result = f"{indent}{self.frame_description(frame)}\n"

            if self.unicode:
                indents = {"├": "├─ ", "│": "│  ", "└": "└─ ", " ": "   "}
            else:
                indents = {"├": "|- ", "│": "|  ", "└": "`- ", " ": "   "}

            if (
                frame.group
                and frame.group.root == frame
                and not self.should_ignore_group(frame.group)
            ):
                result += f"{child_indent}   {self.group_description(frame.group)}"
                for key in indents:
                    indents[key] = "      "
        else:
            result = ""
            indents = {"├": "", "│": "", "└": "", " ": ""}

        if frame.children:
            children_to_be_rendered_indices = [
                i for i, f in enumerate(frame.children) if self.should_render_frame(f)
            ]
            last_rendered_child_index = (
                children_to_be_rendered_indices[-1] if children_to_be_rendered_indices else -1
            )

            for i, child in enumerate(frame.children):
                if i < last_rendered_child_index:
                    c_indent = child_indent + indents["├"]
                    cc_indent = child_indent + indents["│"]
                else:
                    c_indent = child_indent + indents["└"]
                    cc_indent = child_indent + indents[" "]
                result += self.render_frame(child, indent=c_indent, child_indent=cc_indent)

        return result

    def render_frame_flat(self, frame: Frame, indent: str) -> str:
        def walk(frame: Frame):
            frame_id_to_time[frame.identifier] = (
                frame_id_to_time.get(frame.identifier, 0) + frame.total_self_time
                if self.flat_time == "self"
                else frame.time
            )

            frame_id_to_frame[frame.identifier] = frame

            for child in frame.children:
                walk(child)

        frame_id_to_time: Dict[str, float] = {}
        frame_id_to_frame: Dict[str, Frame] = {}

        walk(frame)

        id_time_pairs: List[Tuple[str, float]] = sorted(
            frame_id_to_time.items(), key=(lambda item: item[1]), reverse=True
        )

        if not self.show_all:
            # remove nodes that represent less than 0.1% of the total time
            id_time_pairs = [
                pair for pair in id_time_pairs if pair[1] / self.root_frame.time > 0.001
            ]

        result = ""

        for frame_id, self_time in id_time_pairs:
            result += self.frame_description(frame_id_to_frame[frame_id], override_time=self_time)
            result += "\n"

        return result

    def frame_description(self, frame: Frame, *, override_time: float | None = None) -> str:
        time = override_time if override_time is not None else frame.time
        time_color = self._ansi_color_for_time(time)

        if self.time == "percent_of_total":
            time_str = f"{self.frame_proportion_of_total_time(time) * 100:.1f}%"
        else:
            time_str = f"{time:.3f}"

        value_str = f"{time_color}{time_str}{self.colors.end}"

        class_name = frame.class_name
        if class_name:
            function_name = f"{class_name}.{frame.function}"
        else:
            function_name = frame.function
        function_color = self._ansi_color_for_name(frame)
        function_str = f"{function_color}{function_name}{self.colors.end}"

        code_position_short = frame.code_position_short()
        if code_position_short:
            code_position_str = f"{self.colors.faint}{code_position_short}{self.colors.end}"
        else:
            code_position_str = ""

        return f"{value_str} {function_str}  {code_position_str}"

    def frame_proportion_of_total_time(self, time: float) -> float:
        if self.root_frame.time == 0:
            return 1
        return time / self.root_frame.time

    def _ansi_color_for_time(self, time: float) -> str:
        proportion_of_total = self.frame_proportion_of_total_time(time)

        if proportion_of_total > 0.6:
            return self.colors.red
        elif proportion_of_total > 0.2:
            return self.colors.yellow
        elif proportion_of_total > 0.05:
            return self.colors.green
        else:
            return self.colors.bright_green + self.colors.faint

    def _ansi_color_for_name(self, frame: Frame) -> str:
        if frame.is_application_code:
            return self.colors.bg_dark_blue_255 + self.colors.white_255
        else:
            return ""

    def default_processors(self) -> ProcessorList:
        return [
            processors.remove_importlib,
            processors.remove_tracebackhide,
            processors.merge_consecutive_self_time,
            processors.aggregate_repeated_calls,
            processors.remove_irrelevant_nodes,
            processors.remove_unnecessary_self_time_nodes,
            processors.remove_first_pyinstrument_frames_processor,
            processors.group_library_frames_processor,
        ]

    class colors_enabled:
        red = "\033[31m"
        green = "\033[32m"
        yellow = "\033[33m"
        blue = "\033[34m"
        cyan = "\033[36m"
        bright_green = "\033[92m"
        white = "\033[37m\033[97m"

        bg_dark_blue_255 = "\033[48;5;24m"
        white_255 = "\033[38;5;15m"

        bold = "\033[1m"
        faint = "\033[2m"

        end = "\033[0m"

    class colors_disabled:
        red = ""
        green = ""
        yellow = ""
        blue = ""
        cyan = ""
        bright_green = ""
        white = ""

        bg_dark_blue_255 = ""
        white_255 = ""

        bold = ""
        faint = ""

        end = ""



from __future__ import annotations

import json
import typing
from typing import Any, Callable

from pyinstrument import processors
from pyinstrument.frame import Frame
from pyinstrument.renderers.base import FrameRenderer, ProcessorList
from pyinstrument.session import Session

# pyright: strict


# note: this file is called jsonrenderer to avoid hiding built-in module 'json'.

encode_str = typing.cast(Callable[[str], str], json.encoder.encode_basestring)  # type: ignore


def encode_bool(a_bool: bool):
    return "true" if a_bool else "false"


class JSONRenderer(FrameRenderer):
    """
    Outputs a tree of JSON, containing processed frames.
    """

    output_file_extension = "json"

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def render_frame(self, frame: Frame | None):
        if frame is None:
            return "null"
        # we don't use the json module because it uses 2x stack frames, so
        # crashes on deep but valid call stacks

        property_decls: list[str] = []
        property_decls.append('"function": %s' % encode_str(frame.function))
        property_decls.append('"file_path_short": %s' % encode_str(frame.file_path_short or ""))
        property_decls.append('"file_path": %s' % encode_str(frame.file_path or ""))
        property_decls.append('"line_no": %d' % (frame.line_no or 0))
        property_decls.append('"time": %f' % frame.time)
        property_decls.append('"await_time": %f' % frame.await_time())
        property_decls.append(
            '"is_application_code": %s' % encode_bool(frame.is_application_code or False)
        )

        # can't use list comprehension here because it uses two stack frames each time.
        children_jsons: list[str] = []
        for child in frame.children:
            children_jsons.append(self.render_frame(child))
        property_decls.append('"children": [%s]' % ",".join(children_jsons))

        if frame.group:
            property_decls.append('"group_id": %s' % encode_str(frame.group.id))

        if frame.class_name:
            property_decls.append('"class_name": %s' % encode_str(frame.class_name))

        return "{%s}" % ",".join(property_decls)

    def render(self, session: Session):
        frame = self.preprocess(session.root_frame())

        property_decls: list[str] = []
        property_decls.append('"start_time": %f' % session.start_time)
        property_decls.append('"duration": %f' % session.duration)
        property_decls.append('"sample_count": %d' % session.sample_count)
        property_decls.append('"target_description": %s' % encode_str(session.target_description))
        property_decls.append('"cpu_time": %f' % session.cpu_time)
        property_decls.append('"root_frame": %s' % self.render_frame(frame))

        return "{%s}\n" % ",".join(property_decls)

    def default_processors(self) -> ProcessorList:
        return [
            processors.remove_importlib,
            processors.remove_tracebackhide,
            processors.merge_consecutive_self_time,
            processors.aggregate_repeated_calls,
            processors.remove_irrelevant_nodes,
            processors.remove_unnecessary_self_time_nodes,
            processors.remove_first_pyinstrument_frames_processor,
            processors.group_library_frames_processor,
        ]



from typing import List, Tuple

# pyright: strict


IDENTIFIER_SEP = "\x00"
ATTRIBUTES_SEP = "\x01"

ATTRIBUTE_MARKER_CLASS_NAME = "c"
ATTRIBUTE_MARKER_LINE_NUMBER = "l"
ATTRIBUTE_MARKER_TRACEBACKHIDE = "h"


def parse_frame_info(frame_info: str) -> Tuple[str, List[str]]:
    """
    Parses a frame_info string, returns a tuple of (identifier, attributes),
    where `identifier` is a unique identifier for this code (e.g. a function
    or method), and `attributes` is a list of invocation-specific attributes
    that were captured at profile-time.
    """

    identifier, _, attributes_str = frame_info.partition(ATTRIBUTES_SEP)

    if not attributes_str:
        return identifier, []

    return identifier, attributes_str.split(ATTRIBUTES_SEP)


def frame_info_get_identifier(frame_info: str) -> str:
    """
    Equivalent to `parse_frame_info(frame_info)[0]`, but faster.
    """
    index = frame_info.find(ATTRIBUTES_SEP)

    if index == -1:
        # no attributes
        return frame_info

    return frame_info[0:index]



#!/usr/bin/env python3


from __future__ import annotations

import glob
import os
import subprocess
import sys
import urllib.parse
from pathlib import Path

import click
from packaging.version import InvalidVersion, Version

import pyinstrument

config = [
    # file path, version find/replace format
    ("setup.py", 'version="{}"'),
    ("pyinstrument/__init__.py", '__version__ = "{}"'),
    ("docs/conf.py", 'release = "{}"'),
]

RED = "\u001b[31m"
GREEN = "\u001b[32m"
OFF = "\u001b[0m"


@click.command()
def bump_version() -> None:
    current_version = pyinstrument.__version__

    try:
        commit_date_str = subprocess.run(
            [
                "git",
                "show",
                "--no-patch",
                "--pretty=format:%ci",
                f"v{current_version}^{{commit}}",
            ],
            check=True,
            capture_output=True,
            encoding="utf8",
        ).stdout
        cd_date, cd_time, cd_tz = commit_date_str.split(" ")

        url_opts = urllib.parse.urlencode({"q": f"is:pr merged:>{cd_date}T{cd_time}{cd_tz}"})
        url = f"https://github.com/joerick/pyinstrument/pulls?{url_opts}"

        print(f"PRs merged since last release:\n  {url}")
        print()
    except subprocess.CalledProcessError as e:
        print(e)
        print("Failed to get previous version tag information.")

    git_changes_result = subprocess.run(["git diff-index --quiet HEAD --"], shell=True)
    repo_has_uncommitted_changes = git_changes_result.returncode != 0

    if repo_has_uncommitted_changes:
        print("error: Uncommitted changes detected.")
        sys.exit(1)

    # fmt: off
    print(              'Current version:', current_version)  # noqa
    new_version = input('    New version: ').strip()
    # fmt: on

    try:
        Version(new_version)
    except InvalidVersion:
        print("error: This version doesn't conform to PEP440")
        print("       https://www.python.org/dev/peps/pep-0440/")
        sys.exit(1)

    actions = []

    for path_pattern, version_pattern in config:
        paths = [Path(p) for p in glob.glob(path_pattern)]

        if not paths:
            print(f"error: Pattern {path_pattern} didn't match any files")
            sys.exit(1)

        find_pattern = version_pattern.format(current_version)
        replace_pattern = version_pattern.format(new_version)
        found_at_least_one_file_needing_update = False

        for path in paths:
            contents = path.read_text(encoding="utf8")
            if find_pattern in contents:
                found_at_least_one_file_needing_update = True
                actions.append(
                    (
                        path,
                        find_pattern,
                        replace_pattern,
                    )
                )

        if not found_at_least_one_file_needing_update:
            print(f'''error: Didn't find any occurrences of "{find_pattern}" in "{path_pattern}"''')
            sys.exit(1)

    print()
    print("Here's the plan:")
    print()

    for action in actions:
        path, find, replace = action
        print(f"{path}  {RED}{find}{OFF} → {GREEN}{replace}{OFF}")

    print(f"Then commit, and tag as v{new_version}")

    answer = input("Proceed? [y/N] ").strip()

    if answer != "y":
        print("Aborted")
        sys.exit(1)

    for path, find, replace in actions:
        contents = path.read_text(encoding="utf8")
        contents = contents.replace(find, replace)
        path.write_text(contents, encoding="utf8")

    print("Files updated. If you want to update the changelog as part of this")
    print("commit, do that now.")
    print()

    while input('Type "done" to continue: ').strip().lower() != "done":
        pass

    subprocess.run(
        [
            "git",
            "commit",
            "--all",
            f"--message=Bump version: v{new_version}",
        ],
        check=True,
    )

    subprocess.run(
        [
            "git",
            "tag",
            "--annotate",
            f"--message=v{new_version}",
            f"v{new_version}",
        ],
        check=True,
    )

    print("Done.")


if __name__ == "__main__":
    os.chdir(Path(__file__).parent.parent.resolve())
    bump_version()



from __future__ import annotations

import marshal
from typing import Any, Dict, Tuple

from pyinstrument import processors
from pyinstrument.frame import Frame
from pyinstrument.renderers.base import FrameRenderer, ProcessorList
from pyinstrument.session import Session

# pyright: strict

FrameKey = Tuple[str, int, str]
CallerValue = Tuple[float, int, float, float]
FrameValue = Tuple[float, int, float, float, Dict[FrameKey, CallerValue]]
StatsDict = Dict[FrameKey, FrameValue]


class PstatsRenderer(FrameRenderer):
    """
    Outputs a marshaled dict, containing processed frames in pstat format,
    suitable for processing by gprof2dot and snakeviz.
    """

    output_file_extension = "pstats"
    output_is_binary = True

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    def frame_key(self, frame: Frame) -> FrameKey:
        return (frame.file_path or "", frame.line_no or 0, frame.function)

    def render_frame(self, frame: Frame | None, stats: StatsDict) -> None:
        if frame is None:
            return

        key = self.frame_key(frame)

        if key not in stats:
            # create a new entry
            # being a statistical profiler, we don't know the exact call time or
            # number of calls, they're stubbed out
            call_time = -1
            number_calls = -1
            total_time = 0
            cumulative_time = 0
            callers: dict[FrameKey, CallerValue] = {}
        else:
            call_time, number_calls, total_time, cumulative_time, callers = stats[key]

        # update the total time and cumulative time
        total_time += frame.total_self_time
        cumulative_time += frame.time

        if frame.parent:
            parent_key = self.frame_key(frame.parent)
            if parent_key not in callers:
                p_call_time = -1
                p_number_calls = -1
                p_total_time = 0
                p_cumulative_time = 0
            else:
                p_call_time, p_number_calls, p_total_time, p_cumulative_time = callers[parent_key]

            p_total_time += frame.total_self_time
            p_cumulative_time += frame.time

            callers[parent_key] = p_call_time, p_number_calls, p_total_time, p_cumulative_time

        stats[key] = (call_time, number_calls, total_time, cumulative_time, callers)

        for child in frame.children:
            if not child.is_synthetic:
                self.render_frame(child, stats)

    def render(self, session: Session):
        frame = self.preprocess(session.root_frame())

        stats: StatsDict = {}
        self.render_frame(frame, stats)

        # marshal.dumps returns bytes, so we need to decode it to a string
        # using surrogateescape
        return marshal.dumps(stats).decode(encoding="utf-8", errors="surrogateescape")

    def default_processors(self) -> ProcessorList:
        return [
            processors.remove_importlib,
            processors.remove_tracebackhide,
            processors.merge_consecutive_self_time,
            processors.aggregate_repeated_calls,
            processors.remove_irrelevant_nodes,
            processors.remove_unnecessary_self_time_nodes,
            processors.remove_first_pyinstrument_frames_processor,
        ]



#!/usr/bin/env python3

import argparse
import os
import shutil
import subprocess
import sys

HTML_RENDERER_DIR = "html_renderer"
JS_BUNDLE = "pyinstrument/renderers/html_resources/app.js"
CSS_BUNDLE = "pyinstrument/renderers/html_resources/app.css"

DOCS_PREVIEW_DIR = "docs/_static/preview"

if __name__ == "__main__":
    # chdir to root of repo
    os.chdir(os.path.dirname(__file__))
    os.chdir("..")

    parser = argparse.ArgumentParser()
    parser.add_argument("--force", action="store_true", help="force a rebuild of the bundle")

    args = parser.parse_args()

    js_source_mtime = 0
    for dirpath, dirnames, filenames in os.walk(HTML_RENDERER_DIR):
        if "node_modules" in dirnames:
            dirnames.remove("node_modules")

        for filename in filenames:
            file = os.path.join(dirpath, filename)
            js_source_mtime = max(js_source_mtime, os.path.getmtime(file))

    js_bundle_is_up_to_date = (
        os.path.exists(JS_BUNDLE) and os.path.getmtime(JS_BUNDLE) >= js_source_mtime
    )

    if js_bundle_is_up_to_date and not args.force:
        print("Bundle up-to-date")
        sys.exit(0)

    if subprocess.call("npm --version", shell=True) != 0:
        raise RuntimeError("npm is required to build the HTML renderer.")

    subprocess.check_call("npm ci", cwd=HTML_RENDERER_DIR, shell=True)
    subprocess.check_call("npm run build", cwd=HTML_RENDERER_DIR, shell=True)

    shutil.copyfile(HTML_RENDERER_DIR + "/dist/pyinstrument-html.iife.js", JS_BUNDLE)
    shutil.copyfile(HTML_RENDERER_DIR + "/dist/style.css", CSS_BUNDLE)

    subprocess.check_call("npm run build -- --mode preview", cwd=HTML_RENDERER_DIR, shell=True)
    shutil.rmtree(DOCS_PREVIEW_DIR, ignore_errors=True)
    shutil.copytree(HTML_RENDERER_DIR + "/dist", DOCS_PREVIEW_DIR)



import os
import sys
from test.util import calculate_frame_tree_times, dummy_session

from pytest import approx

from pyinstrument import processors
from pyinstrument.frame import SELF_TIME_FRAME_IDENTIFIER, Frame

ALL_PROCESSORS = [
    processors.aggregate_repeated_calls,
    processors.group_library_frames_processor,
    processors.merge_consecutive_self_time,
    processors.remove_importlib,
    processors.remove_tracebackhide,
    processors.remove_unnecessary_self_time_nodes,
    processors.remove_irrelevant_nodes,
]


def self_time_frame(time):
    return Frame(SELF_TIME_FRAME_IDENTIFIER, time=time)


def test_frame_passthrough_none():
    for processor in ALL_PROCESSORS:
        assert processor(None, options={}) is None


def test_remove_importlib():
    frame = Frame(
        identifier_or_frame_info="<module>\x00sympy/__init__.py\x0012",
        children=[
            Frame(
                identifier_or_frame_info="_handle_fromlist\x00../<frozen importlib._bootstrap>\x00997",
                children=[
                    self_time_frame(0.15),
                    Frame(
                        identifier_or_frame_info="_find_and_load\x00../<frozen importlib._bootstrap>\x00997",
                        children=[
                            self_time_frame(0.05),
                            Frame(
                                identifier_or_frame_info="<module>\x00sympy/polys/polyfuncs.py\x001",
                                children=[self_time_frame(0.05)],
                            ),
                            Frame(
                                identifier_or_frame_info="<module>\x00sympy/polys/partfrac.py\x001",
                                children=[self_time_frame(0.2)],
                            ),
                        ],
                    ),
                    Frame(
                        identifier_or_frame_info="<module>\x00sympy/polys/numberfields.py\x001",
                        children=[self_time_frame(0.05)],
                    ),
                ],
            )
        ],
    )

    calculate_frame_tree_times(frame)
    frame.self_check()

    assert frame.total_self_time == 0.0
    assert frame.time == approx(0.5)

    frame = processors.remove_importlib(frame, options={})
    assert frame
    frame.self_check()

    assert frame.total_self_time == approx(0.2)  # the root gets the self_time from the importlib
    assert frame.time == approx(0.5)
    assert len(frame.children) == 5
    assert frame.children[0].identifier == SELF_TIME_FRAME_IDENTIFIER
    assert frame.children[0].time == 0.15
    assert frame.children[1].identifier == SELF_TIME_FRAME_IDENTIFIER
    assert frame.children[1].time == 0.05

    assert frame.children[2].file_path == "sympy/polys/polyfuncs.py"
    assert frame.children[3].file_path == "sympy/polys/partfrac.py"
    assert frame.children[4].file_path == "sympy/polys/numberfields.py"


def test_remove_hidden():
    frame = Frame(
        identifier_or_frame_info="<module>\x00sympy/__init__.py\x0012\x01h0",
        children=[
            Frame(
                identifier_or_frame_info="_handle_fromlist\x00../foo.py\x00997\x01h1",
                children=[
                    self_time_frame(0.15),
                    Frame(
                        identifier_or_frame_info="_find_and_load\x00../foo.py\x00997\x01h1",
                        children=[
                            self_time_frame(0.05),
                            Frame(
                                identifier_or_frame_info="<module>\x00sympy/polys/polyfuncs.py\x001\x01h0",
                                children=[self_time_frame(0.05)],
                            ),
                            Frame(
                                identifier_or_frame_info="<module>\x00sympy/polys/partfrac.py\x001\x01h0",
                                children=[self_time_frame(0.2)],
                            ),
                        ],
                    ),
                    Frame(
                        identifier_or_frame_info="<module>\x00sympy/polys/numberfields.py\x001\x01h0",
                        children=[self_time_frame(0.05)],
                    ),
                ],
            )
        ],
    )

    calculate_frame_tree_times(frame)
    frame.self_check()

    assert frame.total_self_time == 0.0
    assert frame.time == approx(0.5)

    frame = processors.remove_tracebackhide(frame, options={})
    assert frame
    frame.self_check()

    assert frame.total_self_time == approx(0.2)  # the root gets the self_time from the importlib
    assert frame.time == approx(0.5)
    assert len(frame.children) == 5
    assert frame.children[0].identifier == SELF_TIME_FRAME_IDENTIFIER
    assert frame.children[0].time == 0.15
    assert frame.children[1].identifier == SELF_TIME_FRAME_IDENTIFIER
    assert frame.children[1].time == 0.05


def test_merge_consecutive_self_time():
    frame = Frame(
        identifier_or_frame_info="<module>\x00cibuildwheel/__init__.py\x0012",
        children=[
            Frame(
                identifier_or_frame_info="strip_newlines\x00cibuildwheel/utils.py\x00997",
                children=[self_time_frame(0.1)],
            ),
            self_time_frame(0.2),
            self_time_frame(0.1),
            Frame(
                identifier_or_frame_info="calculate_metrics\x00cibuildwheel/utils.py\x007",
                children=[self_time_frame(0.1)],
            ),
            self_time_frame(0.05),
        ],
    )

    calculate_frame_tree_times(frame)
    frame.self_check()

    assert frame.time == approx(0.55)

    frame = processors.merge_consecutive_self_time(frame, options={})
    assert frame
    frame.self_check()

    assert frame.time == approx(0.55)
    assert len(frame.children) == 4
    assert frame.children[0].total_self_time == approx(0.1)
    assert frame.children[1].identifier == SELF_TIME_FRAME_IDENTIFIER
    assert frame.children[1].time == approx(0.3)
    assert frame.children[2].time == approx(0.1)
    assert frame.children[3].time == approx(0.05)
    assert frame.children[3].identifier == SELF_TIME_FRAME_IDENTIFIER


def test_aggregate_repeated_calls():
    frame = Frame(
        identifier_or_frame_info="<module>\x00cibuildwheel/__init__.py\x0012",
        children=[
            Frame(
                identifier_or_frame_info="strip_newlines\x00cibuildwheel/utils.py\x00997",
                children=[
                    self_time_frame(0.1),
                    Frame(
                        identifier_or_frame_info="scan_string\x00cibuildwheel/utils.py\x0054",
                        children=[self_time_frame(0.2)],
                    ),
                ],
            ),
            self_time_frame(0.1),
            Frame(
                identifier_or_frame_info="strip_newlines\x00cibuildwheel/utils.py\x00997",
                children=[self_time_frame(0.05)],
            ),
            self_time_frame(0.2),
            Frame(
                identifier_or_frame_info="calculate_metrics\x00cibuildwheel/utils.py\x007",
                children=[self_time_frame(0.1)],
            ),
            self_time_frame(0.05),
        ],
    )

    calculate_frame_tree_times(frame)
    frame.self_check()

    assert frame.time == approx(0.8)

    frame = processors.aggregate_repeated_calls(frame, options={})

    assert frame
    frame.self_check()
    assert frame.time == approx(0.8)
    # children should be sorted by time
    assert len(frame.children) == 3
    assert frame.children[0].function == "strip_newlines"
    assert frame.children[0].time == approx(0.35)
    assert frame.children[0].children[0].function == "scan_string"
    assert frame.children[1].identifier == SELF_TIME_FRAME_IDENTIFIER
    assert frame.children[1].time == approx(0.35)
    assert frame.children[2].function == "calculate_metrics"


def test_remove_irrelevant_nodes():
    frame = Frame(
        identifier_or_frame_info="<module>\x00cibuildwheel/__init__.py\x0012",
        children=[
            Frame(
                identifier_or_frame_info="strip_newlines\x00cibuildwheel/utils.py\x00997",
                children=[
                    Frame(
                        identifier_or_frame_info="scan_string\x00cibuildwheel/utils.py\x0054",
                        children=[self_time_frame(10)],
                    ),
                ],
            ),
            self_time_frame(0.5),
            Frame(
                identifier_or_frame_info="strip_newlines\x00cibuildwheel/utils.py\x00997",
                children=[self_time_frame(0.5)],
            ),
            Frame(
                identifier_or_frame_info="calculate_metrics\x00cibuildwheel/utils.py\x007",
                children=[self_time_frame(0.01)],
            ),
        ],
    )

    calculate_frame_tree_times(frame)
    frame.self_check()

    assert frame.time == approx(11.01)

    frame = processors.remove_irrelevant_nodes(frame, options={})

    assert frame
    frame.self_check()
    assert frame.time == approx(11.01)
    # check the calculate metrics function was deleted
    assert len(frame.children) == 3
    assert "calculate_metrics" not in [f.function for f in frame.children]


def test_remove_unnecessary_self_time_nodes():
    frame = Frame(
        identifier_or_frame_info="<module>\x00cibuildwheel/__init__.py\x0012",
        children=[
            Frame(
                identifier_or_frame_info="strip_newlines\x00cibuildwheel/utils.py\x00997",
                children=[
                    self_time_frame(0.2),
                ],
            ),
            self_time_frame(0.5),
            Frame(
                identifier_or_frame_info="strip_newlines\x00cibuildwheel/utils.py\x00997",
                children=[self_time_frame(0.5)],
            ),
            Frame(
                identifier_or_frame_info="calculate_metrics\x00cibuildwheel/utils.py\x007",
                children=[self_time_frame(0.1)],
            ),
        ],
    )
    calculate_frame_tree_times(frame)
    frame.self_check()

    assert frame.time == approx(1.3)

    frame = processors.remove_unnecessary_self_time_nodes(frame, options={})

    assert frame
    frame.self_check()
    assert frame.time == approx(1.3)
    assert len(frame.children) == 4
    # check the self time node was deleted
    strip_newlines_frame = frame.children[0]
    assert strip_newlines_frame.function == "strip_newlines"
    assert len(strip_newlines_frame.children) == 0
    # check the time was absorbed into the parent
    assert strip_newlines_frame.absorbed_time == 0.2
    assert strip_newlines_frame.time == 0.2


def test_group_library_frames_processor(monkeypatch):
    session = dummy_session()
    session.sys_path = ["env/lib/python3.6"]
    session.sys_prefixes = ["env"]
    frame = Frame(
        identifier_or_frame_info="<module>\x00cibuildwheel/__init__.py\x0012",
        children=[
            Frame(
                identifier_or_frame_info="library_function\x00env/lib/python3.6/django/__init__.py\x00997",
                children=[
                    Frame(
                        identifier_or_frame_info="library_inner\x00env/lib/python3.6/django/http.py\x0054",
                        children=[
                            Frame(
                                identifier_or_frame_info="library_callback\x00env/lib/python3.6/django/views.py\x0054",
                                children=[
                                    Frame(
                                        identifier_or_frame_info="<module>\x00cibuildwheel/views.py\x0012",
                                        children=[self_time_frame(0.3)],
                                    ),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
            self_time_frame(0.5),
            Frame(
                identifier_or_frame_info="strip_newlines\x00cibuildwheel/utils.py\x00997",
                children=[self_time_frame(0.5)],
            ),
            Frame(
                identifier_or_frame_info="calculate_metrics\x00cibuildwheel/utils.py\x007",
                children=[self_time_frame(0.1)],
            ),
        ],
        context=session,
    )

    calculate_frame_tree_times(frame)
    frame.self_check()

    assert frame.time == approx(1.4)

    frame = processors.group_library_frames_processor(frame, options={})

    assert frame
    frame.self_check()
    assert frame.time == approx(1.4)
    group_root = frame.children[0]

    group = group_root.group
    assert group
    assert group.root == group_root

    for frame in group.frames:
        assert frame.group == group

    assert group_root in group.frames
    assert group_root.children[0] in group.frames
    assert group_root.children[0].children[0] in group.frames
    assert group_root.children[0].children[0] in group.exit_frames
    assert group_root.children[0].children[0].children[0] not in group.frames



import threading
import time
from test.fake_time_util import fake_time

import pytest

from pyinstrument import Profiler

from .util import do_nothing


def test_profiler_access_from_multiple_threads():
    profiler = Profiler()

    profiler.start()

    thread_exception = None

    def helper():
        while profiler._active_session and len(profiler._active_session.frame_records) < 10:
            time.sleep(0.0001)

        try:
            profiler.stop()
        except Exception as e:
            nonlocal thread_exception
            thread_exception = e

    t1 = threading.Thread(target=helper)
    t1.start()

    while t1.is_alive():
        do_nothing()
    t1.join()

    with pytest.raises(Exception) as excinfo:
        profiler.output_html()

    assert "this profiler is still running" in excinfo.value.args[0]

    assert thread_exception is not None
    assert (
        "Failed to stop profiling. Make sure that you start/stop profiling on the same thread."
        in thread_exception.args[0]
    )

    # the above stop failed. actually stop the profiler
    profiler.stop()



from __future__ import annotations

import json
import os
import sys
from collections import deque
from typing import Any

from pyinstrument.frame import Frame
from pyinstrument.frame_info import frame_info_get_identifier
from pyinstrument.frame_ops import FrameRecordType, build_frame_tree
from pyinstrument.typing import PathOrStr

# pyright: strict


ASSERTION_MESSAGE = (
    "Please raise an issue at https://github.com/joerick/pyinstrument/issues and "
    "let me know how you caused this error!"
)


class Session:
    def __init__(
        self,
        frame_records: list[FrameRecordType],
        start_time: float,
        duration: float,
        min_interval: float,
        max_interval: float,
        sample_count: int,
        start_call_stack: list[str],
        target_description: str,
        cpu_time: float,
        sys_path: list[str],
        sys_prefixes: list[str],
    ):
        """Session()

        Represents a profile session, contains the data collected during a profile session.

        :meta private:
        """
        self.frame_records = frame_records
        self.start_time = start_time
        self.duration = duration
        self.min_interval = min_interval
        self.max_interval = max_interval
        self.sample_count = sample_count
        self.start_call_stack = start_call_stack
        self.target_description = target_description
        self.cpu_time = cpu_time
        self.sys_path = sys_path
        self.sys_prefixes = sys_prefixes
        self._short_file_path_cache = {}

    @staticmethod
    def load(filename: PathOrStr) -> Session:
        """
        Load a previously saved session from disk.

        :param filename: The path to load from.
        :rtype: Session
        """
        with open(filename) as f:
            return Session.from_json(json.load(f))

    def save(self, filename: PathOrStr) -> None:
        """
        Saves a Session object to disk, in a JSON format.

        :param filename: The path to save to. Using the ``.pyisession`` extension is recommended.
        """
        with open(filename, "w") as f:
            json.dump(self.to_json(), f)

    def to_json(self, include_frame_records: bool = True):
        result: dict[str, Any] = {
            "start_time": self.start_time,
            "duration": self.duration,
            "min_interval": self.min_interval,
            "max_interval": self.max_interval,
            "sample_count": self.sample_count,
            "start_call_stack": self.start_call_stack,
            "target_description": self.target_description,
            "cpu_time": self.cpu_time,
            "sys_path": self.sys_path,
            "sys_prefixes": self.sys_prefixes,
        }

        if include_frame_records:
            result["frame_records"] = self.frame_records

        return result

    @staticmethod
    def from_json(json_dict: dict[str, Any]):
        return Session(
            frame_records=json_dict["frame_records"],
            start_time=json_dict["start_time"],
            min_interval=json_dict.get("min_interval", 0.001),
            max_interval=json_dict.get("max_interval", 0.001),
            duration=json_dict["duration"],
            sample_count=json_dict["sample_count"],
            start_call_stack=json_dict["start_call_stack"],
            target_description=json_dict["target_description"],
            cpu_time=json_dict["cpu_time"] or 0,
            sys_path=json_dict.get("sys_path", sys.path),
            sys_prefixes=json_dict.get("sys_prefixes", Session.current_sys_prefixes()),
        )

    @staticmethod
    def combine(session1: Session, session2: Session) -> Session:
        """
        Combines two :class:`Session` objects.

        Sessions that are joined in this way probably shouldn't be interpreted
        as timelines, because the samples are simply concatenated. But
        aggregate views (the default) of this data will work.

        :rtype: Session
        """
        if session1.start_time > session2.start_time:
            # swap them around so that session1 is the first one
            session1, session2 = session2, session1

        return Session(
            frame_records=session1.frame_records + session2.frame_records,
            start_time=session1.start_time,
            min_interval=min(session1.min_interval, session2.min_interval),
            max_interval=max(session1.max_interval, session2.max_interval),
            duration=session1.duration + session2.duration,
            sample_count=session1.sample_count + session2.sample_count,
            start_call_stack=session1.start_call_stack,
            target_description=session1.target_description,
            cpu_time=session1.cpu_time + session2.cpu_time,
            sys_path=(
                session1.sys_path + [p for p in session2.sys_path if p not in session1.sys_path]
            ),
            sys_prefixes=list(set([*session1.sys_prefixes, *session2.sys_prefixes])),
        )

    @staticmethod
    def current_sys_prefixes() -> list[str]:
        return [sys.prefix, sys.base_prefix, sys.exec_prefix, sys.base_exec_prefix]

    def root_frame(self, trim_stem: bool = True) -> Frame | None:
        """
        Parses the internal frame records and returns a tree of :class:`Frame`
        objects. This object can be rendered using a :class:`Renderer`
        object.

        :rtype: A :class:`Frame` object, or None if the session is empty.
        """
        root_frame = build_frame_tree(self.frame_records, context=self)

        if root_frame is None:
            return None

        if trim_stem:
            root_frame = self._trim_stem(root_frame)

        return root_frame

    def _trim_stem(self, frame: Frame):
        # trim the start of the tree before any branches.
        # we also don't want to trim beyond the call to profiler.start()

        start_stack = deque(frame_info_get_identifier(info) for info in self.start_call_stack)

        if start_stack.popleft() != frame.identifier:
            # the frame doesn't match where the profiler was started. Don't trim.
            return frame

        while frame.total_self_time == 0 and len(frame.children) == 1:
            # check child matches the start_call_stack, otherwise stop descending
            if len(start_stack) == 0 or frame.children[0].identifier != start_stack.popleft():
                break

            frame = frame.children[0]

        frame.remove_from_parent()
        return frame

    _short_file_path_cache: dict[str, str]

    def shorten_path(self, path: str) -> str:
        """
        Shorten a path to a more readable form, relative to sys_path. Used by
        Frame.short_file_path.
        """
        if path in self._short_file_path_cache:
            return self._short_file_path_cache[path]

        result = path
        # if os.sep doesn't appear, probably not a file path at all, more
        # likely <built-in> or similar
        if len(path.split(os.sep)) > 1:
            for sys_path_entry in self.sys_path:
                # On Windows, if path and sys_path_entry are on
                # different drives, relpath will result in exception,
                # because it cannot compute a relpath in this case.
                # The root cause is that on Windows, there is no root
                # dir like '/' on Linux.
                try:
                    candidate = os.path.relpath(path, sys_path_entry)
                except ValueError:
                    continue

                if not result or (len(candidate.split(os.sep)) < len(result.split(os.sep))):
                    result = candidate

        self._short_file_path_cache[path] = result

        return result



from __future__ import annotations

import contextvars
import time
from typing import Any

import pytest

from ..util import busy_wait
from .util import parametrize_setstatprofile


@parametrize_setstatprofile
def test_context_type(setstatprofile):
    with pytest.raises(TypeError):
        setstatprofile(lambda f, e, a: 0, 1e6, "not a context var")
        setstatprofile(None)


profiler_context_var: contextvars.ContextVar[object | None] = contextvars.ContextVar(
    "profiler_context_var", default=None
)


@parametrize_setstatprofile
def test_context_tracking(setstatprofile):
    profile_calls = []

    def profile_callback(frame, event, arg):
        nonlocal profile_calls
        profile_calls.append((frame, event, arg))

    profiler_1 = object()
    profiler_2 = object()

    context_1 = contextvars.copy_context()
    context_2 = contextvars.copy_context()

    context_1.run(profiler_context_var.set, profiler_1)
    context_2.run(profiler_context_var.set, profiler_2)

    setstatprofile(
        profile_callback,
        1e10,  # set large interval so we only get context_change events
        profiler_context_var,
    )

    context_1.run(busy_wait, 0.001)
    context_2.run(busy_wait, 0.001)

    setstatprofile(None)

    assert all(c[1] == "context_changed" for c in profile_calls)
    assert len(profile_calls) == 4

    new, old, _ = profile_calls[0][2]
    assert old is None
    assert new is profiler_1

    new, old, _ = profile_calls[1][2]
    assert old is profiler_1
    assert new is None

    new, old, _ = profile_calls[2][2]
    assert old is None
    assert new is profiler_2

    new, old, _ = profile_calls[3][2]
    assert old is profiler_2
    assert new is None



import asyncio
import dataclasses
import inspect
import json
import time
from functools import partial
from test.fake_time_util import fake_time
from typing import Generator, Optional

import pytest

from pyinstrument import Profiler, renderers
from pyinstrument.frame import Frame
from pyinstrument.renderers.speedscope import SpeedscopeEvent, SpeedscopeEventType, SpeedscopeFrame
from pyinstrument.session import Session

from .util import assert_never, busy_wait, flaky_in_ci

# Utilities #


def long_function_a():
    time.sleep(0.25)


def long_function_b():
    time.sleep(0.5)


class ClassWithMethods:
    def long_method(self):
        time.sleep(0.25)

    @staticmethod
    def long_static_method():
        time.sleep(0.25)

    @classmethod
    def long_class_method(cls):
        time.sleep(0.25)


# Tests #


def test_collapses_multiple_calls_by_default():
    profiler = Profiler()

    with fake_time():
        profiler.start()

        long_function_a()
        long_function_b()
        long_function_a()
        long_function_b()

        profiler.stop()

    text_output = profiler.output_text()
    print(text_output)

    # output should be something like:
    # 1.500 test_collapses_multiple_calls_by_default  test/test_profiler.py:25
    # |- 0.500 long_function_a  test/test_profiler.py:17
    # |- 0.500 long_function_b  test/test_profiler.py:20

    assert text_output.count("1.500 test_collapses_multiple_calls_by_default") == 1
    assert text_output.count("0.500 long_function_a") == 1
    assert text_output.count("1.000 long_function_b") == 1


def test_profiler_retains_multiple_calls():
    profiler = Profiler()

    with fake_time():
        profiler.start()

        long_function_a()
        long_function_b()
        long_function_a()
        long_function_b()

        profiler.stop()

    print(profiler.output_text())

    assert profiler.last_session
    frame = profiler.last_session.root_frame()
    assert frame
    assert frame.function == "test_profiler_retains_multiple_calls"
    assert len(frame.children) == 4


def test_two_functions():
    profiler = Profiler()

    with fake_time():
        profiler.start()

        long_function_a()
        long_function_b()

        profiler.stop()

    print(profiler.output_text())

    assert profiler.last_session

    frame = profiler.last_session.root_frame()

    assert frame
    assert frame.function == "test_two_functions"
    assert len(frame.children) == 2

    frame_b, frame_a = sorted(frame.children, key=lambda f: f.time, reverse=True)

    assert frame_a.function == "long_function_a"
    assert frame_b.function == "long_function_b"

    # busy CI runners can be slow to wake up from the sleep. So we relax the
    # ranges a bit
    assert frame_a.time == pytest.approx(0.25, abs=0.1)
    assert frame_b.time == pytest.approx(0.5, abs=0.2)


def test_class_methods():
    profiler = Profiler()

    with fake_time():
        profiler.start()

        obj = ClassWithMethods()
        obj.long_method()
        obj.long_class_method()
        obj.long_static_method()
        partial(obj.long_method)()

        profiler.stop()

    text_output = profiler.output_text()
    print(text_output)

    # output should be something like:
    # 1.000 test_class_methods  test/test_profiler.py:124
    # |- 0.500 ClassWithMethods.long_method  test/test_profiler.py:29
    # |  `- 0.500 FakeClock.sleep  test/fake_time_util.py:19
    # |- 0.250 ClassWithMethods.long_class_method  test/test_profiler.py:36
    # |  `- 0.250 FakeClock.sleep  test/fake_time_util.py:19
    # `- 0.250 long_static_method  test/test_profiler.py:32
    #    `- 0.250 FakeClock.sleep  test/fake_time_util.py:19
    assert text_output.count("0.500 ClassWithMethods.long_method") == 1
    assert text_output.count("0.250 ClassWithMethods.long_class_method") == 1
    assert text_output.count("0.250 long_static_method") == 1


def test_context_manager():
    with fake_time():
        with Profiler() as profiler:
            long_function_a()
            long_function_b()

    assert profiler.last_session
    frame = profiler.last_session.root_frame()
    assert frame
    assert frame.function == "test_context_manager"
    assert len(frame.children) == 2


def test_json_output():
    with fake_time():
        with Profiler() as profiler:
            long_function_a()
            long_function_b()

    output_data = profiler.output(renderers.JSONRenderer())

    output = json.loads(output_data)
    assert "root_frame" in output

    root_frame = output["root_frame"]

    assert root_frame["function"] == "test_json_output"
    assert len(root_frame["children"]) == 2


def test_speedscope_output():
    with fake_time():
        with Profiler() as profiler:
            long_function_a()
            long_function_b()

    output_data = profiler.output(renderers.SpeedscopeRenderer())

    output = json.loads(output_data)

    file_level_schema_fields = {
        "$schema",
        "name",
        "exporter",
        "activeProfileIndex",
        "profiles",
        "shared",
    }
    for file_field in file_level_schema_fields:
        assert file_field in output

    assert output["$schema"] == "https://www.speedscope.app/file-format-schema.json"
    assert "pyinstrument" in output["exporter"]
    assert output["activeProfileIndex"] is None
    assert "CPU profile" in output["name"]

    assert "frames" in output["shared"]
    speedscope_frame_list = output["shared"]["frames"]

    # Distinct functions called stores indices in key-value pairs because function
    # index lookup needed for Speedscope event list tests. Were we not testing
    # the event list, the distinct functions called could be stored as a tuple.
    distinct_functions_called = {
        "test_speedscope_output": 0,
        "long_function_a": 1,
        "sleep": 2,
        "long_function_b": 3,
    }
    assert len(speedscope_frame_list) == len(distinct_functions_called)
    speedscope_frame_fields = tuple(field.name for field in dataclasses.fields(SpeedscopeFrame))
    for function_name, frame_index in distinct_functions_called.items():
        for frame_field in speedscope_frame_fields:
            assert frame_field in speedscope_frame_list[frame_index]
        assert speedscope_frame_list[frame_index]["name"] == function_name

    speedscope_profile_list = output["profiles"]
    assert len(speedscope_profile_list) == 1
    speedscope_profile = speedscope_profile_list[0]
    speedscope_profile_fields = ("type", "name", "unit", "startValue", "endValue", "events")
    for profile_field in speedscope_profile_fields:
        assert profile_field in speedscope_profile

    # speedscope_profile["endValue"] is not tested because a fake_time mock
    # timer is used to replace time.time, and this mock causes session.duration
    # to differ from self._event_time just before exiting SpeedscopeRenderer.render
    assert speedscope_profile["type"] == "evented"
    assert speedscope_profile["unit"] == "seconds"
    assert speedscope_profile["startValue"] == 0.0

    output_event_tuple = (
        SpeedscopeEvent(
            SpeedscopeEventType.OPEN,
            0.0,
            distinct_functions_called["test_speedscope_output"],
        ),
        SpeedscopeEvent(
            SpeedscopeEventType.OPEN,
            0.0,
            distinct_functions_called["long_function_a"],
        ),
        SpeedscopeEvent(
            SpeedscopeEventType.OPEN,
            0.0,
            distinct_functions_called["sleep"],
        ),
        SpeedscopeEvent(
            SpeedscopeEventType.CLOSE,
            0.25,
            distinct_functions_called["sleep"],
        ),
        SpeedscopeEvent(
            SpeedscopeEventType.CLOSE,
            0.25,
            distinct_functions_called["long_function_a"],
        ),
        SpeedscopeEvent(
            SpeedscopeEventType.OPEN,
            0.25,
            distinct_functions_called["long_function_b"],
        ),
        SpeedscopeEvent(
            SpeedscopeEventType.OPEN,
            0.25,
            distinct_functions_called["sleep"],
        ),
        SpeedscopeEvent(
            SpeedscopeEventType.CLOSE,
            0.75,
            distinct_functions_called["sleep"],
        ),
        SpeedscopeEvent(
            SpeedscopeEventType.CLOSE,
            0.75,
            distinct_functions_called["long_function_b"],
        ),
        SpeedscopeEvent(
            SpeedscopeEventType.CLOSE,
            0.75,
            distinct_functions_called["test_speedscope_output"],
        ),
    )

    speedscope_event_list = speedscope_profile["events"]
    assert len(speedscope_event_list) == len(output_event_tuple)
    speedscope_event_fields = tuple(field.name for field in dataclasses.fields(SpeedscopeEvent))
    for event_index, speedscope_event in enumerate(speedscope_event_list):
        for event_field in speedscope_event_fields:
            assert event_field in speedscope_event

        assert speedscope_event["type"] == output_event_tuple[event_index].type.value
        assert speedscope_event["frame"] == output_event_tuple[event_index].frame
        assert speedscope_event["at"] == pytest.approx(output_event_tuple[event_index].at)


def test_empty_profile():
    with Profiler() as profiler:
        pass
    profiler.output(renderer=renderers.ConsoleRenderer())


@flaky_in_ci
def test_state_management():
    profiler = Profiler()

    assert profiler.last_session is None
    assert profiler.is_running == False

    profiler.start()

    assert profiler.last_session is None
    assert profiler.is_running == True

    busy_wait(0.1)

    profiler.stop()

    assert profiler.is_running == False
    assert profiler.last_session is not None
    assert profiler.last_session.duration == pytest.approx(0.1, rel=0.2)

    # test a second session, does it merge with the first?
    profiler.start()

    assert profiler.is_running == True
    busy_wait(0.1)

    profiler.stop()

    assert profiler.is_running == False

    assert profiler.last_session is not None
    assert profiler.last_session.duration == pytest.approx(0.2, rel=0.2)

    # test a reset
    profiler.reset()
    assert profiler.last_session is None

    # test a reset while running
    profiler.start()
    assert profiler.is_running == True
    profiler.reset()
    assert profiler.is_running == False
    assert profiler.last_session is None


@pytest.mark.parametrize(
    "profiler_method_name,renderer_class",
    [
        ("print", renderers.ConsoleRenderer),
        ("output_text", renderers.ConsoleRenderer),
        ("output_html", renderers.HTMLRenderer),
    ],
)
def test_profiler_convenience_methods_have_all_options_available(
    profiler_method_name, renderer_class
):
    method = getattr(Profiler, profiler_method_name)

    method_signature = inspect.signature(method)
    renderer_signature = inspect.signature(renderer_class)

    for name, parameter in renderer_signature.parameters.items():
        if name == "self":
            continue

        if profiler_method_name == "output_html" and name in {"timeline", "show_all"}:
            # these options have been deprecated
            continue

        assert (
            name in method_signature.parameters
        ), f"Parameter {name} is missing from Profiler.{profiler_method_name}. {parameter}"
        method_parameter = method_signature.parameters[name]

        if profiler_method_name == "print" and name in {"color", "unicode"}:
            # print has a mechanism of autodetecting these
            continue
        assert (
            method_parameter.default == parameter.default
        ), f"Parameter {name} has a different default value in Profiler.{profiler_method_name}. {parameter}"
        assert (
            method_parameter.annotation == parameter.annotation
        ), f"Parameter {name} has a different annotation in Profiler.{profiler_method_name}. {parameter}"



import threading
import time

current_time = 0.0

subscriber_lock = threading.Lock()
update_lock = threading.Lock()
thread_should_exit = False
thread_alive = False


class Subscription:
    def __init__(self, interval: float, id: int):
        self.interval = interval
        self.id = id


subscribers = []


def get_interval(max_interval: float):
    if subscribers:
        return min(sub.interval for sub in subscribers)
    return max_interval


def timing_thread():
    global current_time, thread_should_exit

    while not thread_should_exit:
        interval = get_interval(1.0)
        acquired = update_lock.acquire(timeout=interval)
        if acquired:
            update_lock.release()
        current_time = time.perf_counter()


def pyi_timing_thread_subscribe(desired_interval: float):
    global thread_alive, thread_should_exit, current_time

    with subscriber_lock:
        if not thread_alive:
            update_lock.acquire()
            thread_should_exit = False
            threading.Thread(target=timing_thread).start()
            thread_alive = True
            current_time = time.perf_counter()

        ids = [sub.id for sub in subscribers]
        new_id = 0
        while new_id in ids:
            new_id += 1

        subscribers.append(Subscription(desired_interval, new_id))

        update_lock.release()
        update_lock.acquire()

    return new_id


def pyi_timing_thread_unsubscribe(id: int):
    with subscriber_lock:
        subscriber_to_remove = next((sub for sub in subscribers if sub.id == id), None)

        if subscriber_to_remove:
            subscribers.remove(subscriber_to_remove)

            if not subscribers:
                global thread_should_exit, thread_alive
                thread_should_exit = True
                update_lock.release()
                thread_alive = False
            return 0
        else:
            raise Exception("PYI_TIMING_THREAD_NOT_SUBSCRIBED")


def pyi_timing_thread_get_time() -> float:
    return current_time


def pyi_timing_thread_get_interval() -> float:
    return get_interval(float("inf")) if thread_alive else -1.0



from pathlib import Path

import pytest

from pyinstrument.__main__ import main
from pyinstrument.renderers.base import FrameRenderer

from .util import BUSY_WAIT_SCRIPT

fake_renderer_instance = None


class FakeRenderer(FrameRenderer):
    def __init__(self, time=None, **kwargs):
        self.time = time
        super().__init__(**kwargs)
        global fake_renderer_instance
        fake_renderer_instance = self
        print("instance")

    def default_processors(self):
        """
        Return a list of processors that this renderer uses by default.
        """
        return []

    def render(self, session) -> str:
        return ""


def test_renderer_option(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    (tmp_path / "test_program.py").write_text(BUSY_WAIT_SCRIPT)
    monkeypatch.setattr(
        "sys.argv",
        [
            "pyinstrument",
            "-r",
            "test.test_cmdline_main.FakeRenderer",
            "-p",
            "time=percent_of_total",
            "test_program.py",
        ],
    )
    monkeypatch.chdir(tmp_path)

    global fake_renderer_instance
    fake_renderer_instance = None

    main()

    assert fake_renderer_instance is not None
    assert fake_renderer_instance.time == "percent_of_total"


def test_json_renderer_option(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    (tmp_path / "test_program.py").write_text(BUSY_WAIT_SCRIPT)
    monkeypatch.setattr(
        "sys.argv",
        [
            "pyinstrument",
            "-r",
            "test.test_cmdline_main.FakeRenderer",
            "-p",
            'processor_options={"some_option": 44}',
            "test_program.py",
        ],
    )
    monkeypatch.chdir(tmp_path)

    global fake_renderer_instance
    fake_renderer_instance = None

    main()

    assert fake_renderer_instance is not None
    assert fake_renderer_instance.processor_options["some_option"] == 44


def test_dotted_renderer_option(monkeypatch: pytest.MonkeyPatch, tmp_path: Path):
    (tmp_path / "test_program.py").write_text(BUSY_WAIT_SCRIPT)
    monkeypatch.setattr(
        "sys.argv",
        [
            "pyinstrument",
            "-r",
            "test.test_cmdline_main.FakeRenderer",
            "-p",
            "processor_options.other_option=13",
            "test_program.py",
        ],
    )
    monkeypatch.chdir(tmp_path)

    global fake_renderer_instance
    fake_renderer_instance = None

    main()

    assert fake_renderer_instance is not None
    assert fake_renderer_instance.processor_options["other_option"] == 13



#!/usr/bin/env python3

import os
import subprocess
import sys
from glob import glob
from pathlib import Path

ROOT_DIR = Path(__file__).parent.parent


def main():
    os.chdir(ROOT_DIR)
    for script in glob(str("examples/demo_scripts/*.py")):
        subprocess.run([sys.executable, "bin/create_sample_json.py", script], check=True)


if __name__ == "__main__":
    main()
